/***************************************************************************
 *   Authors:                                                              *
 *   - Nicolas Brodu                                                       *
 *                                                                         *
 *   Copyrights:                                                           *
 *   - INRIA                                                               *
 *   - Nicolas Brodu (initial base code, some modifications in SVN)        *
 *                                                                         *
 *   This file is free software; you can redistribute it and/or modify     *
 *   it under the terms of the GNU General Public License as               *
 *   published by the Free Software Foundation; either version 2 of the    *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU Lesser General Public License for more details.                   *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 ***************************************************************************/

#include <iostream>

#include <math.h>
#include <assert.h>

#include "AdaBoost.h"

using namespace std;
using namespace itpp;


AdaBoost::AdaBoost(ClassifierFactory* weakClassifierFactory, int numBoostingSteps, int resampleSize, int forceResampleFlag) :
factory(weakClassifierFactory),
nsteps(numBoostingSteps),
resample(resampleSize),
classifiers(numBoostingSteps),
combiWeights(numBoostingSteps)
{
    assert(nsteps>0);
    for (int i=0; i<nsteps; ++i) classifiers[i] = factory->instanciateClassifier();

    if (classifiers[0]->supportWeightedInstance()) {
        if (!forceResampleFlag) resample = 0;
        else if (!resample) resample = -1; // forced resampling but using training set size explicitly
    } else if (!resample) {
        resample = -1;
    }
}

AdaBoost::~AdaBoost() {
    for (int i=0; i<nsteps; ++i) delete classifiers[i];
}


//////////////////////////////////////////////////////////////////////////////
// Options
//////////////////////////////////////////////////////////////////////////////

std::string AdaBoost::getName() {
    return "Adaboost: " + classifiers[0]->getName();
}

bool AdaBoost::parseOption(int option, std::string option_arg) {
    bool ret = false;
    for (int i=0; i<nsteps; ++i) ret |= classifiers[i]->parseOption(option,option_arg);
    return ret;
}

//////////////////////////////////////////////////////////////////////////////
// classifier work
//////////////////////////////////////////////////////////////////////////////

bool AdaBoost::setVFoldCrossValidation(int v) {
    bool ret = false;
    for (int i=0; i<nsteps; ++i) ret |= classifiers[i]->setVFoldCrossValidation(v);
    return ret;
}

double AdaBoost::finishTraining() {

    int ntraininstances = trainSet.size();
    if (!ntraininstances) return 0.0;

    int tsize = resample>0?resample:ntraininstances;
    vector<int> tidx(tsize); // ==0 or -1 both use train set size
    // No resampling => all train instances are used
    if (!resample) for (int i=0; i<ntraininstances; ++i) tidx[i] = i;

    // initial value is equi distribution
    // normalization is done later on, use 1.0 here for "max relative influence"
    // 1.0 is a value with definite meaning independently of the training set size. See classifier.h
    vector<double> weights(ntraininstances, 1.0);
    double sumOfWeights = ntraininstances;
    vector<double> currentErrors(ntraininstances, 1.0);

    // re-init combination
    sumCombiWeights = 0.0;

    double epsilon_t = 0.0;

    for (int adabooststep = 0; adabooststep < nsteps; ++adabooststep) {
        // resampling occurs each step if used
        if (resample) for (int i=0; i<tsize; ++i) tidx[i] = drawRandomInstance();

        // First part: train using the current weights
        classifiers[adabooststep]->prepareForTraining(tsize, trainSet[0].size());
        for (int i=0; i<tsize; ++i) classifiers[adabooststep]->train(trainSet[tidx[i]],trainClasses[tidx[i]],weights[tidx[i]]);
        classifiers[adabooststep]->finishTraining();

        // Part 2: compute weighted training error on the real set
        epsilon_t = 0.0;            // weighting gives more influence to badly classified instances => might go above 0.5 error
	double classif_error = 0.0; // unweighted classif error
        for (int i=0; i<ntraininstances; ++i) {
            // Error in 0..1
            currentErrors[i] = 0.5 * abs(trainClasses[i] - classifiers[adabooststep]->predict(trainSet[i]));
            epsilon_t += weights[i] * currentErrors[i];
            classif_error += currentErrors[i];
        }
        epsilon_t /= sumOfWeights; // normalize to [0..1]
        if (classif_error/ntraininstances >= 0.5) {
            if (verbose) { cout << "Adaboost warning: classifier is more than 50% wrong." << endl; }
            //return 0.0; // really no confidence in this result...
            // Should test on effective train set with tsize?
        }
        // beta might be null if the classifier is perfect
        // epsilon_t in [0..0.5] => beta in [0..1]
        double beta = epsilon_t / (1.0 - epsilon_t);

        // Update the weights for next iteration
        sumOfWeights = 0.0;
        double maxW = 0.0;
        for (int i=0; i<ntraininstances; ++i) {
            // If beta is null then currentErrors is too, and so pow = 1.0 and weight is non null
            // beta in 0..1, (1.0 - currentErrors[i]) in 0..1, => pow in 0..1
            weights[i] *= pow(beta, 1.0 - currentErrors[i]);
            if (weights[i]>maxW) maxW=weights[i];
            sumOfWeights += weights[i];
        }
        // renormalize so that max weight is 1.0 and we keep the meaning "max relative influence"
        for (int i=0; i<ntraininstances; ++i) weights[i] /= maxW;
        sumOfWeights /= maxW;

        // Give a global weight for this step classifier
        // direct application might result in infinite weight!
        combiWeights[adabooststep] = -log(beta);
        if (!isnormal(combiWeights[adabooststep])) {
            // Consider that since instance weights are unchanged by beta = 0, then so does the classifier weight
            if (adabooststep>0) combiWeights[adabooststep] = combiWeights[adabooststep-1];
            // first step jackpot!!! start with combiweight = 1 and it will be copied since beta will be the same with the same instance weights!
            else combiWeights[0] = 1.0;
        }
        sumCombiWeights += combiWeights[adabooststep];

    }

    // Use the last step error as a confidence measure for this classifier
    return 1.0 - epsilon_t;

}

double AdaBoost::predict(const vec& instance) {

    double cumulatedPrediction = 0.0;
    for (int adabooststep = 0; adabooststep < nsteps; ++adabooststep) {
        cumulatedPrediction += classifiers[adabooststep]->predict(instance) * combiWeights[adabooststep];
    }
    return cumulatedPrediction / sumCombiWeights;
}




