/***************************************************************************
 *   Authors:                                                              *
 *   - Nicolas Brodu                                                       *
 *                                                                         *
 *   Copyrights:                                                           *
 *   - INRIA                                                               *
 *                                                                         *
 *   This file is free software; you can redistribute it and/or modify     *
 *   it under the terms of the GNU General Public License as               *
 *   published by the Free Software Foundation; either version 2 of the    *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU Lesser General Public License for more details.                   *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 ***************************************************************************/

#ifndef CLASSIFIER_INTERFACE
#define CLASSIFIER_INTERFACE

/*

NOTE: There is SOME degree of abstraction, but there are still assumptions present in the main file. In particular the main file knows all possible factories.

This is because the purpose is here to just organize the code so as to make it more maintainable and reusable, the purpose is NOT to make an ultra-generic classifier library. There would be many other problems to solve for an hypothetical ultra-generic classifier system anyway, and no time to solve them (ex: stable ABI and API, and most importantly enough interest from a user base).

As it is now it is easy to extract the classifier classes in order to use them in another application, which is already quite nice.

*/

#include <vector>
#include <string>
#include <ostream>

#include <itpp/itbase.h>

// Child classes are real classifier
struct ClassifierInterface {

    ClassifierInterface();
    virtual ~ClassifierInterface();

    // Set v-fold cross-validation mode, if relevant
    // v has already been checked to be a valid number, 0 < v < ninstances
    // This function is called at most once, before any train/predict cycles
    // return true if supported, false if not
    virtual bool setVFoldCrossValidation(int v);

    // Each training instance is provided with a weight that tells the relative importance of that instance (possible all equals).
    // The classifier may use the weight information, in which case it should indicate so here by returning true (default is false).
    // When the weight information is ignored, Adaboost will work by resampling the training set, so the relative frequencies
    // of the instances in the new training set correspond approximately to their weights. In that case all weights will be set to 1.0.
    // When the classifier is able to use the weight information directly Adaboost will rely on it and provide the training set as is.
    // Weights will have values between 0.0 and 1.0, it is the classifier responsibility to normalize them properly by their sum if needed.
    // This convention allows a caller to use 0.0 for "don't use this instance" and 1.0 for "this instance has the largest influence",
    // irrespectively of the number of instances.
    virtual bool supportWeightedInstance();

    //This method is called just before a new training session
    // with the number of instances that will be provided and the dimension of an instance vector
    // The default implementation reserve memory for storing all instances
    // You do not need to call this implementation if you process instances only once,
    // but you might find calling it useful if you need to store the instances (ex: for cross-validation)
    virtual void prepareForTraining(int ninstances, int instancedim);

    // This method is called once per training instance
    // classification is either +1 or -1
    // The instance weight gives the importance of this instance, between 0 and 1. The default weight is 1.
    // The default implementation stores the instance for later use
    // You do not need to call this implementation if you process instances only once,
    // on the other hand if all you do here is storing the instance then you do not need to overload this method
    virtual void train(const itpp::vec& instance, double classification, double instanceWeight = 1.0);

    // This method is called once all training instances have been fed
    // The classifier should perform v-fold cross-validation at this point if this option was set
    // Returns a confidence value that indicates whether the classifier thinks it will generalize well
    // This value is between 0 and 1, use 1 if this feature is unsupported by this classifier (default is to be confident in its own result!)
    virtual double finishTraining();

    // This method is called after a whole training session, once for every test instance
    // It must return a double, with sign indicating the class -1/+1. Values between -1 and 1 allows to express confidence levels.
    virtual double predict(const itpp::vec& instance) = 0;

    // First argument as returned by getopt, second as a copy of optarg
    // return true if option was handled by this method
    // Note: All options here are pre-parsed by the main file (S) syntax, so you should
    // only need to take care of the simple, normal, syntax. ex: -X 1,2,3
    // When a (S) syntax is used, each intance here gets its specific option as if it was global (G)
    // Consequently characters : and ; are forbidden as separators in the parsing.
    virtual bool parseOption(int option, std::string option_arg);

    virtual std::string getName();

    // accessors do what they are intended to do by default, but may be overloaded
    // Useful for method wrappers like adaboost, in order to propagate/broadcast to wrapped classifiers.
    virtual bool getQuiet();
    virtual void setQuiet(bool q);
    virtual bool getVerbose();
    virtual void setVerbose(bool v);

protected:
    bool quiet;
    bool verbose;

    // Provide a facility for child classifiers these are empty and may be unused
    // A classifier may decide (or not) to store all instances, for example for cross-validation in finishTraining
    // These fields are managed by the default implementations of prepareForTraining and train
    std::vector<itpp::vec> trainSet;
    std::vector<double> trainClasses;
    std::vector<double> trainWeights;
    // maintained by the train function
    std::vector<double> cumulatedWeights;

    // Facility to draw an instance at random from the training set according to its probability as given by the weights
    // returns the index of the instance in the above vectors
    // The Random() method from the Utility class is used, so you can set the random seed reproducibly
    virtual int drawRandomInstance();

};

// Provide meta-info and classifier creation on demand
struct ClassifierFactory {
    ClassifierFactory();
    virtual ~ClassifierFactory();

    enum ClassifierInputType {
        FeatureVector,
        DistributionSample
    };

    virtual ClassifierInputType getType();

    // Create an instance of a specialized Classifier
    // Will be delete'd when unused
    virtual ClassifierInterface* instanciateClassifier() = 0;

    // print an help text in the given output stream -- no need to create a classifier at this point
    virtual void printClassifierHelpText(std::ostream& os);

    // options accepted as in getopt -- please use Capital letters -- no need to create a classifier at this point
    virtual std::string getClassifierOptions();

    // Return a number used as a method selection ID in the options
    virtual int getMethodId() = 0;
};

#endif

