/***************************************************************************
 *   Authors:                                                              *
 *   - Nicolas Brodu                                                       *
 *                                                                         *
 *   Copyrights:                                                           *
 *   - INRIA                                                               *
 *   - Nicolas Brodu (initial base code, some modifications in SVN)        *
 *                                                                         *
 *   This file is free software; you can redistribute it and/or modify     *
 *   it under the terms of the GNU General Public License as               *
 *   published by the Free Software Foundation; either version 2 of the    *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU Lesser General Public License for more details.                   *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 ***************************************************************************/

#include <iostream>
#include <set>

#include <stdlib.h>

#include "HiddenLayerPerceptron.h"

using namespace std;
using namespace itpp;

static const double defaultLearningRate = 0.07;
static const int defaultNbSteps = 500;


ClassifierInterface* HiddenLayerPerceptronFactory::instanciateClassifier() {
    return new HiddenLayerPerceptronClassifier();
}

HiddenLayerPerceptronClassifier::HiddenLayerPerceptronClassifier() {
    nhidden = nhiddenMaxCV = nhiddenStepCV = crossValidation = 0;
    nbpsteps = defaultNbSteps;
    learningRate = defaultLearningRate;
    resample = 0;
}

HiddenLayerPerceptronClassifier::~HiddenLayerPerceptronClassifier() {
}


//////////////////////////////////////////////////////////////////////////////
// Options
//////////////////////////////////////////////////////////////////////////////


ClassifierFactory::ClassifierInputType HiddenLayerPerceptronFactory::getType() {
    return FeatureVector;
}

std::string HiddenLayerPerceptronClassifier::getName() {
     return "Multi-layer Perceptron with one hidden layer";
}

int HiddenLayerPerceptronFactory::getMethodId() {
    return 3;
}

std::string HiddenLayerPerceptronFactory::getClassifierOptions() {
    return "H:L:S:R:";
}

void HiddenLayerPerceptronFactory::printClassifierHelpText(ostream& os) {
os<<
"    3: Multi-layer Perceptron with one hidden layer (w)\n"
"      -H <int>: (*) Number of hidden neurons (default=ninputs*1.2)\n"
"      -L <float>: Learning rate for the gradient descent after back-propagation (default="<<defaultLearningRate<<").\n"
"      -S <int>: Number of learning steps for the gradient descent (default="<<defaultNbSteps<<").\n"
"      -R <int>: Resample N training instances with probabilities given by their weights at each gradient descent step (default = weight the error function)\n"
;
}

bool HiddenLayerPerceptronClassifier::parseOption(int option, std::string option_arg) {
    switch(option) {
        case 'H': {
            for (unsigned int i=0; i<option_arg.size(); ++i) if (option_arg[i]==',') option_arg[i]=' ';
            stringstream tokenizer(option_arg);
            tokenizer >> nhidden;
            tokenizer >> nhiddenMaxCV;
            tokenizer >> nhiddenStepCV;
            if ((nhidden<1) || (nhiddenStepCV<0) || (nhiddenStepCV && (nhiddenMaxCV<nhidden))) {
                cerr << "Invalid number of hidden neurons: " << optarg << endl;
                exit(1);
            }
            return true;
        }
        case 'L': {
            learningRate = atof(option_arg.c_str());
            if (learningRate<=0.0) {
                cerr << "Invalid learning rate: " << option_arg << endl;
                exit(1);
            }
            return true;
        }
        case 'S': {
            nbpsteps = atoi(option_arg.c_str());
            if (nbpsteps<1) {
                cerr << "Invalid number of gradient descent steps: " << option_arg << endl;
                exit(1);
            }
            return true;
        }
        case 'R': {
            resample = atoi(option_arg.c_str());
            if (resample<1) {
                cerr << "Invalid number of training instances to resample: " << option_arg << endl;
                exit(1);
            }
            return true;
        }
    }
    return false;
}


//////////////////////////////////////////////////////////////////////////////
// classifier work
//////////////////////////////////////////////////////////////////////////////

bool HiddenLayerPerceptronClassifier::setVFoldCrossValidation(int v) {
    crossValidation = v;
    return true;
}

bool HiddenLayerPerceptronClassifier::supportWeightedInstance() {
    return true; // whether through resampling or through weighted error function
}

double HiddenLayerPerceptronClassifier::finishTraining() {

    int ntraininstances = trainSet.size();
    if (!ntraininstances) return 0.0;

    double maxW = 0.0;
    for (int i=0; i<ntraininstances; ++i) if (maxW<trainWeights[i]) maxW=trainWeights[i];
    if (maxW==0.0) maxW = 1.0; // avoid 0/0 NaN

    int ninputs = trainSet[0].size();
    // use categorical data mapping for non-linear classifier
    double output[2], gradout[2];

    double bestPerf = -1;
    if (crossValidation && ((nhiddenStepCV!=0) || (nhidden==0))) {

        int resampCVSetSize = 0;
        if (resample) resampCVSetSize = (resample + crossValidation - 1) / crossValidation;

        if (verbose) cout << "Cross-validating number of hidden neurons" << endl;
        vector<int> subsetBounds(crossValidation+1);
        for (int subset = 0; subset < crossValidation; ++subset) {
            subsetBounds[subset] = subset * ntraininstances / crossValidation;
        }
        subsetBounds[crossValidation] = ntraininstances; // end limit, after array
        set<int> cvH;
        if (nhiddenStepCV) {
            int rangeH = nhiddenMaxCV - nhidden;
            if (nhiddenStepCV>1) for (int ih = 0; ih < nhiddenStepCV; ++ih) {
                cvH.insert(nhidden + (ih * rangeH) / (nhiddenStepCV - 1));
            }
            else cvH.insert((nhidden + nhiddenMaxCV) / 2);
        } else {
            int hmin = ninputs/5; if (hmin<1) hmin = 1;
            int rangeH = ninputs*5 - hmin;
            for (int ih = 0; ih <= 10; ++ih) {
                cvH.insert(hmin + (ih * rangeH) / 10);
            }
        }

        int bestH = *cvH.begin();
        for (set<int>::const_iterator hit = cvH.begin(); hit!=cvH.end(); ++hit) {
            int nhiddencv = *hit;
            if (verbose) cout << "nhidden=" << nhiddencv << flush;
            double perf = 0.0;
            for (int subset = 0; subset < crossValidation; ++subset) {
                mlp = MultiLayerPerceptron (ninputs,nhiddencv,2);
                // Gradient descent
                for (int bpstep = 0; bpstep < nbpsteps; ++bpstep) {

                    vector<int> instanceIdx;
                    if (resample) {
                        for (int i=0; i<resampCVSetSize; ++i) {
                            // get random index out of exclusion range
                            int idx;
                            do {
                                idx = drawRandomInstance();
                            } while (idx>=subsetBounds[subset] && idx<subsetBounds[subset+1]);
                            instanceIdx.push_back(idx);
                        }
                    }
                    else {
                        for (int i=0; i<subsetBounds[subset]; ++i) instanceIdx.push_back(i);
                        for (int i=subsetBounds[subset+1]; i<ntraininstances; ++i) instanceIdx.push_back(i);
                    }

                    for (unsigned int i=0; i<instanceIdx.size(); ++i) {
                        int idx = instanceIdx[i];
//                    for (int i=0; i<ntraininstances; ++i) {
//                        if (i>=subsetBounds[subset] && i<subsetBounds[subset+1]) continue;

                        const double* x = trainSet[idx]._data();
                        mlp.computeOutput(x, output);
                        gradout[0] = output[0] - trainClasses[idx];
                        gradout[1] = output[1] + trainClasses[idx];
                        if (!resample) {gradout[0] *= trainWeights[idx]/maxW; gradout[1] *= trainWeights[idx]/maxW;}
                        if (i==0) mlp.backPropagate(x, output, gradout);
                        else mlp.batchBackPropagateAccumulate(x, output, gradout);
                    }
//                    mlp.batchBackPropagateTerminate(ntraininstances - (subsetBounds[subset+1] - subsetBounds[subset]));
                    mlp.batchBackPropagateTerminate(instanceIdx.size());
                    mlp.learn(learningRate);
                }
                // Add performance on the subset that was left out
                for (int i=subsetBounds[subset]; i<subsetBounds[subset+1]; ++i) {
                    mlp.computeOutput(trainSet[i]._data(), output);
                    double cl = output[0] - output[1];
                    if ((cl * trainClasses[i]) > 0.0) perf += trainWeights[i];
                }
            }
            // favor minimum number of hidden units for which perf is the same so as to help generalisation: use > and not >=
            // no need to regularize weight total for the comparison
            if (perf>bestPerf) {
                bestPerf = perf;
                bestH = nhiddencv;
            }
            // now got cross-validation performance on the whole training set
            if (verbose) cout << ", perf=" << (perf*100.0/cumulatedWeights.back()) << ", best perf=" << (bestPerf*100.0/cumulatedWeights.back()) << endl;
        }
        if (!quiet) cout << "Cross-validation selected " << bestH << " hidden units (for " << ninputs << " inputs)" <<endl;
        mlp = MultiLayerPerceptron (ninputs,bestH,2);
    } else {
        if (nhidden) mlp = MultiLayerPerceptron (ninputs,nhidden,2);
        else mlp = MultiLayerPerceptron (ninputs,5 * ninputs / 4, 2);
    }

    // Gradient descent on selected number of hidden units + whole training set
    int effectiveN = ntraininstances;
    if (resample) effectiveN = resample;
    for (int bpstep = 0; bpstep < nbpsteps; ++bpstep) {
        for (int i=0; i<effectiveN; ++i) {
            int idx = resample ? drawRandomInstance() : i;
            const double* x = trainSet[idx]._data();
            mlp.computeOutput(x, output);
            gradout[0] = output[0] - trainClasses[idx];
            gradout[1] = output[1] + trainClasses[idx];
            if (!resample) {gradout[0] *= trainWeights[idx]/maxW; gradout[1] *= trainWeights[idx]/maxW;}
            if (i==0) mlp.backPropagate(x, output, gradout);
            else mlp.batchBackPropagateAccumulate(x, output, gradout);
        }
        mlp.batchBackPropagateTerminate(effectiveN);
        mlp.learn(learningRate);
    }

    // network is ready to serve!

    // Use cross-validated performance as generalization ability estimate
    if (bestPerf!=-1) return bestPerf/cumulatedWeights.back();

    // Otherwise use weighted classification performance of the network
    double clperf = 0.0;
    for (int i=0; i<ntraininstances; ++i) {
        mlp.computeOutput(trainSet[i]._data(), output);
        double cl = output[0] - output[1];
        if ((cl * trainClasses[i]) > 0.0) clperf += trainWeights[i];
    }
    return clperf / cumulatedWeights.back();
}

double HiddenLayerPerceptronClassifier::predict(const vec& instance) {
    double output[2];
    mlp.computeOutput(instance._data(), output);
    return (output[0] - output[1]) * 0.5;
}

