/***************************************************************************
 *   Authors:                                                              *
 *   - Nicolas Brodu                                                       *
 *                                                                         *
 *   Copyrights:                                                           *
 *   - INRIA                                                               *
 *   - Nicolas Brodu (initial base code, some modifications in SVN)        *
 *                                                                         *
 *   This file is free software; you can redistribute it and/or modify     *
 *   it under the terms of the GNU General Public License as               *
 *   published by the Free Software Foundation; either version 2 of the    *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU Lesser General Public License for more details.                   *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 ***************************************************************************/

#include <math.h>

#include <iostream>

#include "LinearFisherDiscriminant.h"

using namespace std;
using namespace itpp;

ClassifierInterface* LinearFisherDiscriminantFactory::instanciateClassifier() {
    return new LinearFisherDiscriminantClassifier();
}

//////////////////////////////////////////////////////////////////////////////
// Options
//////////////////////////////////////////////////////////////////////////////


ClassifierFactory::ClassifierInputType LinearFisherDiscriminantFactory::getType() {
    return FeatureVector;
}

std::string LinearFisherDiscriminantClassifier::getName() {
    return "Linear Fisher discriminant with Gaussian threshold";
}

int LinearFisherDiscriminantFactory::getMethodId() {
    return 2;
}

void LinearFisherDiscriminantFactory::printClassifierHelpText(ostream& os) {
    os << "    2: Linear Fisher discriminant with Gaussian threshold\n";
}


//////////////////////////////////////////////////////////////////////////////
// classifier work
//////////////////////////////////////////////////////////////////////////////

void LinearFisherDiscriminantClassifier::prepareForTraining(int ninstances, int instancedim) {
    mean.resize(2);
    cov.resize(2);
    gw.set_size(2);
    for (int c=0; c<2; ++c) {
        mean[c].set_size(instancedim);
        mean[c].zeros();
        cov[c].set_size(instancedim,instancedim);
        cov[c].zeros();
    }
    gw.zeros();
    this->ninstances = ninstances;
}

void LinearFisherDiscriminantClassifier::train(const itpp::vec& instance, double classification, double /*instanceWeight: IGNORED TODO*/) {
    int classnum = ((int)classification+1)/2; // -1/1 => 0/1
    mean[classnum] += instance;
    cov[classnum] += instance * instance.transpose();
    ++gw(classnum);
}

double LinearFisherDiscriminantClassifier::finishTraining() {
    for (int c=0; c<2; ++c) {
        mean[c] /= gw(c);
        // sample covariance => n-1
        double invNm1 = 1.0;
        if (gw(c)>1) invNm1 /= gw(c) - 1.0;
        cov[c] = (cov[c] - gw(c) * mean[c] * mean[c].transpose()) * invNm1;
        // convert counts to weights
        gw(c) /= ninstances;
    }

    // Now we can have within-class scatter matrix
    mat within_scatter = gw(0) * cov[0] + gw(1) * cov[1];
    // store projection vector
    vec meandiff = mean[0] - mean[1];
    discriminantsW = meandiff.transpose() * inv(within_scatter);

    // threshold?
    double m0 = dot(discriminantsW, mean[0]);
    double m1 = dot(discriminantsW, mean[1]);
    //discriminantsT = (m0 + m1)*0.5;

    // each class projected variance
    mat mv0 = discriminantsW.transpose() * cov[0] * discriminantsW;
    mat mv1 = discriminantsW.transpose() * cov[1] * discriminantsW;
    double v0 = mv0(0,0);
    double v1 = mv1(0,0);

    // Take the gaussians intersection between the class centers, if it exists
    double a = 1.0 / v0 - 1.0 / v1;
    double b = -2.0 * (m0 / v0 - m1 / v1);
    double c = (m0 * m0 / v0) - (m1 * m1 / v1) + log(v0) - log(v1);
    double sqdelta = b * b - 4 * a * c;
    if (sqdelta>=0.0) {
        sqdelta = sqrt(sqdelta);
        //double s1 = (-b - sqrt(delta)) / (2.0 * a);
        //double s2 = (-b + sqrt(delta)) / (2.0 * a);
        //if (!quiet) cout << "s1=" << s1 << ", s2=" << s2 << endl;
        //if (!quiet) cout << "m0=" << m0 << ", m1=" << m1 << endl;
        discriminantsT = (-b -sqdelta) / (2.0 * a);
    }
    else {
        discriminantsT = (m0 + m1)*0.5;
    }

    // weight of this classifier = fisher discriminant ratio
    double classif_weight = (m0 - m1) * (m0 - m1) / (v0 + v1);
    if (verbose) cout << "FDR=" << classif_weight;

    // 0-variance => if means are separated: perfect, otherwise: bad
    if (!isnormal(classif_weight)) {
        if (m0==m1) classif_weight = 0.0;
        else classif_weight = 1.0;
    }
    // Otherwise squash the FDR into 0..1 using a sigmoid (see below)
    else classif_weight /= 1.0 + classif_weight;

    if (verbose) cout << ", w=" << discriminantsW << ", b=" << discriminantsT << ", classif_weight=" << classif_weight << endl;

    return classif_weight;
}

double LinearFisherDiscriminantClassifier::predict(const itpp::vec& instance) {
    double res = discriminantsT - dot(discriminantsW, instance);
    // Use a sigmoid here as well
    return res / (1.0 + fabs(res));
}


