/***************************************************************************
 *   Authors:                                                              *
 *   - Nicolas Brodu                                                       *
 *                                                                         *
 *   Copyrights:                                                           *
 *   - INRIA                                                               *
 *   - Nicolas Brodu (initial base code, some modifications in SVN)        *
 *                                                                         *
 *   This file is free software; you can redistribute it and/or modify     *
 *   it under the terms of the GNU General Public License as               *
 *   published by the Free Software Foundation; either version 2 of the    *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU Lesser General Public License for more details.                   *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 ***************************************************************************/

#include <iostream>
#include <math.h>

#include "LinearMinSumSquaredError.h"

using namespace std;
using namespace itpp;

ClassifierInterface* LinearMinSumSquaredErrorFactory::instanciateClassifier() {
    return new LinearMinSumSquaredErrorClassifier();
}

//////////////////////////////////////////////////////////////////////////////
// Options
//////////////////////////////////////////////////////////////////////////////


ClassifierFactory::ClassifierInputType LinearMinSumSquaredErrorFactory::getType() {
    return FeatureVector;
}

std::string LinearMinSumSquaredErrorClassifier::getName() {
    return "Decision hyperplane with min sum of squared error";
}

int LinearMinSumSquaredErrorFactory::getMethodId() {
    return 1;
}

void LinearMinSumSquaredErrorFactory::printClassifierHelpText(ostream& os) {
    os << "    1: Linear min sum of squared error (w)\n";
}

//////////////////////////////////////////////////////////////////////////////
// classifier work
//////////////////////////////////////////////////////////////////////////////

bool LinearMinSumSquaredErrorClassifier::supportWeightedInstance() {
    return true;
}

void LinearMinSumSquaredErrorClassifier::prepareForTraining(int ninstances, int instancedim) {
    linear_regression_matrix.set_size(ninstances, instancedim+1); // allow constant b in Y=aX+b
    linear_regression_target.set_size(ninstances, 1); // +1 / -1 for linear combinations need no categorical mapping
    curInstance = 0;
}

void LinearMinSumSquaredErrorClassifier::train(const itpp::vec& instance, double classification, double instanceWeight) {
    // null weight corresponds to null influence.
    linear_regression_matrix.set_submatrix(curInstance, curInstance, 0, instance.size()-1, instanceWeight * instance.transpose() );
    linear_regression_matrix(curInstance,instance.size()) = instanceWeight;
    linear_regression_target(curInstance) = classification * instanceWeight;
    ++curInstance;
}

double LinearMinSumSquaredErrorClassifier::finishTraining() {

    regweights = backslash(linear_regression_matrix, linear_regression_target);

    vec delta = (linear_regression_matrix * regweights - linear_regression_target).get_col(0);
    double error = 0.0;
    for (int i=0; i<delta.size(); ++i) error += delta(i) * delta(i);
    error = sqrt(error / delta.size());

    if (verbose) cout << "regweights: " << regweights.transpose() << ", " << "RMSE: " << error << endl;
    error = 1.0 - error;
    return error>0.0?error:0.0;
}

double LinearMinSumSquaredErrorClassifier::predict(const itpp::vec& instance) {
    vec instance_const(instance.size()+1);
    for (int i = 0; i<instance.size(); ++i) instance_const(i) = instance(i);
    instance_const(instance.size()) = 1;
    mat resm = instance_const.transpose() * regweights;
    double res = resm(0,0);
    if (res>1.0) res = 1.0;
    if (res<-1.0) res = -1.0;
    return res;
}

