/***************************************************************************
 *   Authors:                                                              *
 *   - Nicolas Brodu                                                       *
 *                                                                         *
 *   Copyrights:                                                           *
 *   - INRIA                                                               *
 *   - Nicolas Brodu (initial base code, some modifications in SVN)        *
 *                                                                         *
 *   This file is free software; you can redistribute it and/or modify     *
 *   it under the terms of the GNU General Public License as               *
 *   published by the Free Software Foundation; either version 2 of the    *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU Lesser General Public License for more details.                   *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 ***************************************************************************/

#include <iostream>

#include "MaximumLikelihoodGaussianMixture.h"

// ITPP v>=4 only
#ifndef USE_ITPP_V3_10
#include <itpp/base/math/log_exp.h>
#else
#include <itpp/base/scalfunc.h>
#endif


using namespace std;
using namespace itpp;

ClassifierInterface* MaximumLikelihoodGaussianMixtureFactory::instanciateClassifier() {
    return new MaximumLikelihoodGaussianMixtureClassifier();
}

MaximumLikelihoodGaussianMixtureFactory::MaximumLikelihoodGaussianMixtureFactory() {
}

MaximumLikelihoodGaussianMixtureFactory::~MaximumLikelihoodGaussianMixtureFactory() {
}

MaximumLikelihoodGaussianMixtureClassifier::MaximumLikelihoodGaussianMixtureClassifier() {
}

MaximumLikelihoodGaussianMixtureClassifier::~MaximumLikelihoodGaussianMixtureClassifier() {
}

//////////////////////////////////////////////////////////////////////////////
// Options
//////////////////////////////////////////////////////////////////////////////

std::string MaximumLikelihoodGaussianMixtureFactory::getClassifierOptions() {
    return "";
}

bool MaximumLikelihoodGaussianMixtureClassifier::parseOption(int option, std::string option_arg) {
    return false;
}

ClassifierFactory::ClassifierInputType MaximumLikelihoodGaussianMixtureFactory::getType() {
    return FeatureVector;
}

std::string MaximumLikelihoodGaussianMixtureClassifier::getName() {
    return "Maximum Likelihood Gaussian mixture";
}

int MaximumLikelihoodGaussianMixtureFactory::getMethodId() {
    return 4;
}

void MaximumLikelihoodGaussianMixtureFactory::printClassifierHelpText(ostream& os) {
os<<"    4: Maximum Likelihood Gaussian mixture (using the Chernoff bound for accumulation)\n";
}


//////////////////////////////////////////////////////////////////////////////
// classifier work
//////////////////////////////////////////////////////////////////////////////

void MaximumLikelihoodGaussianMixtureClassifier::prepareForTraining(int ninstances, int instancedim) {
    mean.resize(2);
    cov_covinv.resize(2);
    gw.set_size(2);
    for (int c=0; c<2; ++c) {
        mean[c].set_size(instancedim);
        mean[c].zeros();
        cov_covinv[c].set_size(instancedim,instancedim);
        cov_covinv[c].zeros();
    }
    gw.zeros();
    this->ninstances = ninstances;
}

void MaximumLikelihoodGaussianMixtureClassifier::train(const itpp::vec& instance, double classification, double instanceWeight) {
    int classnum = ((int)classification+1)/2; // -1/1 => 0/1
    mean[classnum] += instance;
    cov_covinv[classnum] += instance * instance.transpose();
    ++gw(classnum);
}

double MaximumLikelihoodGaussianMixtureClassifier::finishTraining() {

    logNormalizingFactor.set_size(2);

    const double minusHalfDimLog2Pi = -0.3990899341790575261 * mean[0].size();
    vec dcov(2);

    for (int c=0; c<2; ++c) {
        mean[c] /= gw(c);
        // sample covariance => n-1
        double invNm1 = 1.0;
        if (gw(c)>1) invNm1 /= gw(c) - 1.0;
        cov_covinv[c] = (cov_covinv[c] - gw(c) * mean[c] * mean[c].transpose()) * invNm1;
        // convert counts to weights
        gw(c) /= ninstances;
        dcov(c) = fabs(det(cov_covinv[c]));
        // precompute normalizing factor from covariance matrix
        logNormalizingFactor(c) = log(gw(c)) + minusHalfDimLog2Pi - 0.5*log(dcov(c));
    }


    // Reuse weighting scheme from Steven Lemm, Christin Sch�fer, and Gabriel Curio / BCI Competition 2003
    // The quantity mentioned as "easily obtained" is the Chernoff Bound using the Bhattacharyya distance.
    mat halfcov = 0.5 * (cov_covinv[0] + cov_covinv[1]);
    mat halfcovinv = inv(halfcov);
    vec mdiff = mean[0] - mean[1];
    mat scalar = mdiff.transpose() * halfcovinv * mdiff;
    double Bhattacharyya_Distance = 0.125 * scalar(0,0) + 0.5 * log( fabs(det(halfcov)) / sqrt( dcov(0)*dcov(1) ) );
    if (verbose) cout << "Bhattacharyya_Distance=" << Bhattacharyya_Distance;
    double Chernoff_Bound = sqrt(gw(0) * gw(1)) * exp(-Bhattacharyya_Distance);
    if (verbose) cout << ", Chernoff_Bound=" << Chernoff_Bound << endl;

    // convert cov to cov_inv for prediction phase
    for (int c=0; c<2; ++c) cov_covinv[c] = inv(cov_covinv[c]);

    double weight = 1.0 - Chernoff_Bound*2.0;
    return (weight>0.0)?weight:0.0; // more robust to numerical errors.
}

double MaximumLikelihoodGaussianMixtureClassifier::classLogLikelihood(const itpp::vec& instance, int classif) {
    vec delta = instance - mean[classif];
    mat scalar = delta.transpose() * (cov_covinv[classif] * delta);
    return logNormalizingFactor(classif) - 0.5 * scalar(0,0);
}

double MaximumLikelihoodGaussianMixtureClassifier::predict(const vec& instance) {

    double l0 = classLogLikelihood(instance, 0);
    double l1 = classLogLikelihood(instance, 1);

    double p0, p1;
    if ((l0>700.0) || (l1>700.0) || (l0<-700.0) || (l1<-700.0)) {
        if (l0==l1) return 0.0;
        if (l0>l1) p0 = 1.0;
        else p0 = 0.0;
        p1 = 1.0 - p0;
    } else {
        p0 = trunc_exp(l0);
        p1 = trunc_exp(l1);
    }

    if (p0>p1) return -p0;
    if (p0<p1) return p1;
    return 0.0;
}
