/***************************************************************************
 *   Authors:                                                              *
 *   - Nicolas Brodu                                                       *
 *                                                                         *
 *   Copyrights:                                                           *
 *   - INRIA                                                               *
 *   - Nicolas Brodu (initial base code, some modifications in SVN)        *
 *                                                                         *
 *   This file is free software; you can redistribute it and/or modify     *
 *   it under the terms of the GNU General Public License as               *
 *   published by the Free Software Foundation; either version 2 of the    *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU Lesser General Public License for more details.                   *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 ***************************************************************************/

#include <iostream>

#include "VariationalBayesianGaussianMixture.h"

using namespace std;
using namespace itpp;

static const int defaultnemstep =50;

ClassifierInterface* VariationalBayesianGaussianMixtureFactory::instanciateClassifier() {
    return new VariationalBayesianGaussianMixtureClassifier();
}

VariationalBayesianGaussianMixtureClassifier::VariationalBayesianGaussianMixtureClassifier() {
    nemstep = defaultnemstep;
}

//////////////////////////////////////////////////////////////////////////////
// Options
//////////////////////////////////////////////////////////////////////////////


ClassifierFactory::ClassifierInputType VariationalBayesianGaussianMixtureFactory::getType() {
    return FeatureVector;
}

std::string VariationalBayesianGaussianMixtureClassifier::getName() {
    return "Variational Bayesian Gaussian Mixture";
}

int VariationalBayesianGaussianMixtureFactory::getMethodId() {
    return 5;
}

std::string VariationalBayesianGaussianMixtureFactory::getClassifierOptions() {
    return "M:";
}

void VariationalBayesianGaussianMixtureFactory::printClassifierHelpText(ostream& os) {
os<<
"    5: Variational Bayesian Gaussian mixture\n"
"      -M <int>: Number of Expectation-Maximisation steps for estimating the hyperparameters (default="<<defaultnemstep<<").\n"
;
}

bool VariationalBayesianGaussianMixtureClassifier::parseOption(int option, std::string option_arg) {
    if (option == 'M') {
        nemstep = atoi(option_arg.c_str());
        if (nemstep<1) {
            cerr << "Invalid number of expectation-maximisation steps: " << optarg << endl;
            exit(1);
        }
        return true;
    }
    return false;
}


//////////////////////////////////////////////////////////////////////////////
// classifier work
//////////////////////////////////////////////////////////////////////////////

bool VariationalBayesianGaussianMixtureClassifier::supportWeightedInstance() {
    return false;
}

void VariationalBayesianGaussianMixtureClassifier::prepareForTraining(int ninstances, int instancedim) {
    ClassifierInterface::prepareForTraining(ninstances, instancedim);
    mean.resize(2);
    sumcov.resize(2);
    classcounts.set_size(2);
    for (int c=0; c<2; ++c) {
        mean[c].set_size(instancedim);
        mean[c].zeros();
        sumcov[c].set_size(instancedim,instancedim);
        sumcov[c].zeros();
    }
    classcounts.zeros();
}

void VariationalBayesianGaussianMixtureClassifier::train(const itpp::vec& instance, double classification, double instanceWeight) {
    ClassifierInterface::train(instance, classification, instanceWeight);
    int classnum = ((int)classification+1)/2; // -1/1 => 0/1
    mean[classnum] += instance;
    sumcov[classnum] += instance * instance.transpose();
    ++classcounts(classnum);
}

double VariationalBayesianGaussianMixtureClassifier::finishTraining() {

    int ntraininstances = trainSet.size();
    if (!ntraininstances) return 0.0;

    int nfeatures = trainSet[0].size();

    hmog.init(2,nfeatures);

    for (int c=0; c<2; ++c) {
        //gw(c) = classcounts(c) / ntraininstances;
        //cout << "gw(c)="<<gw(c)<<endl;
        mean[c] /= classcounts(c);
        sumcov[c] -= (classcounts(c) * mean[c]) * mean[c].transpose();

        // Wishart distribution prior parameters (on the inverse sample covariance matrix)
        hmog.S(c) = sumcov[c];//inv(sumcov[c]);

        hmog.v(c) = classcounts(c) - 1; // degrees of freedom
        // Dirichlet prior: observation counts + 1
        hmog.lambda(c) = classcounts(c) + 1;

        // Normal distribution prior (for the sample mean)
        hmog.m(c) = mean[c];
        hmog.beta(c) = classcounts(c); // variance scale factor from sample mean to mean of mean
    }

    // take into account the initial parameters
    hmog.prepareForEstimation();

    double classif = 0.5, prevclassif=0.5;

    // Now perform VB EM steps
    Array<vec> posteriorProba(ntraininstances);
    for (int i=0; i<ntraininstances; ++i) posteriorProba(i).set_size(2);
    vec classProbaSum(2);

    int emstep = 0;

    // nemstep is set by the options
    for (; emstep<nemstep; ++emstep) {

        // experimental
//        double datalh = 0.0;

        bool nanerror = false;
        // VB E-step: update posterior proba
        for (int i=0; i<ntraininstances; ++i) {
            const vec& x = trainSet[i];
            double normalization_Factor = 0.0;
            for (int g=0; g<2; ++g) {
                posteriorProba(i)(g) = hmog.estimatePosteriorLogLikelihood(g, x);
                if (isnan(posteriorProba(i)(g))) nanerror = true;
                // clamp logs below maximum representable number to avoid +/-inf problems
                if ((posteriorProba(i)(g)) > 700.0) posteriorProba(i)(g) = 700.0;
                if ((posteriorProba(i)(g)) < -700.0) posteriorProba(i)(g) = -700.0;
//if (g==((int)(trainClasses[i])+1)/2) datalh -= posteriorProba(i)(g);
                posteriorProba(i)(g) = trunc_exp(posteriorProba(i)(g));
                normalization_Factor += posteriorProba(i)(g);
                if (isnan(normalization_Factor)) nanerror = true;
            }
            // Check for numerical stability issues...
            if (nanerror) break;

            for (int g=0; g<2; ++g) posteriorProba(i)(g) /= normalization_Factor;
        }
        // stop EM here and stick with previous step if error could not be recovered
        if (nanerror) break;

        // VB M-step: Update parameters: means
        classProbaSum.zeros();
        for (int g=0; g<2; ++g) mean[g].zeros();
        for (int i=0; i<ntraininstances; ++i) {
            const vec& x = trainSet[i];
            classProbaSum += posteriorProba(i);
            for (int g=0; g<2; ++g) mean[g] += x * posteriorProba(i)(g);
        }
        for (int g=0; g<2; ++g) hmog.m(g) = hmog.m(g) * hmog.beta(g) + mean[g];

        for (int g=0; g<2; ++g) mean[g] /= classProbaSum(g);

        for (int g=0; g<2; ++g) hmog.m(g) /= hmog.beta(g) + classProbaSum(g);
        // VB M-step: Update parameters: hyper-parameters
        hmog.beta += classProbaSum;
        hmog.v += classProbaSum;
        hmog.lambda += classProbaSum;

        // VB M-step: Update parameters: covariance of means dist matrix
        for (int g=0; g<2; ++g) {
            for (int i=0; i<ntraininstances; ++i) {
                const vec& x = trainSet[i];
                vec xdelta = x - mean[g];
                hmog.S(g) += posteriorProba(i)(g) * xdelta * xdelta.transpose();
            }
            vec mdelta = mean[g] - hmog.m(g);
            double factor = hmog.beta(g) * classProbaSum(g) / (hmog.beta(g) + classProbaSum(g));
            hmog.S(g) += factor * mdelta * mdelta.transpose();
        }

        // take into account the new parameters
        hmog.prepareForEstimation();

        // TODO: detect when likelihood of the data set does not increase anymore
        // PB: max reached -700!!!
//cout << datalh << endl;
//    }
    
    
        // Using Student's t distribution, as in Hagai Attias original paper.
        // regularization done by baysian approach. Now assume the model generalizes well : weight = classification success, weighted by x proba
        Array<MultivariateStudentT> mst(2);
        for (int g=0; g<2; ++g) {
            mst(g).dof = (int)hmog.v(g) + 1 - hmog.D;
            mst(g).mean = hmog.m(g);
            double factor = (hmog.beta(g) + 1.0) / (hmog.beta(g) * mst(g).dof);
            mst(g).precision = inv(factor * hmog.S(g));
            mst(g).prepareForEstimation();
        }
        vec mixing(2); double mixsum = 0.0;
        for (int g=0; g<2; ++g) mixsum += hmog.lambda(g);
        for (int g=0; g<2; ++g) mixing(g) = hmog.lambda(g) / mixsum;
    
        classif = 0.0;
        double spx = 0.0;
        for (int i=0; i<ntraininstances; ++i) {
            const vec& x = trainSet[i];
            // density estimation: compute p(x)
            double px = 0.0;
            for (int g=0; g<2; ++g) px += mixing(g) * mst(g).density(x);
            spx += px;
            // compute p(C|x) for each class C
            double l0 = hmog.estimatePosteriorLogLikelihood(0,x);
            double l1 = hmog.estimatePosteriorLogLikelihood(1,x);
            double p0, p1;
            if ((l0>700.0) || (l1>700.0) || (l0<-700.0) || (l1<-700.0)) {
                if (l0>l1) p0 = 1.0;
                else p0 = 0.0;
                p1 = 1.0 - p0;
            } else {
                p0 = trunc_exp(l0);
                p1 = trunc_exp(l1);
                p0 /= p0 + p1;
                p1 = 1.0 - p0;
            }
            classif += ((trainClasses[i]<0.0) ? p0 : p1) * px;
        }

        classif /= spx;

        // Arbitrarily stop EM when change is <epsilon
        if (fabs(classif - prevclassif) < 1e-6) break;
        prevclassif = classif;

    }

    if (verbose) {
        if (emstep==nemstep) cout << "VariationalBayesianGaussianMixtureClassifier: max number of steps reached" << endl;
        else cout << "VariationalBayesianGaussianMixtureClassifier: convergence reached" << endl;
        if (verbose) cout << "VariationalBayesianGaussianMixtureClassifier: weight=" << classif << endl;
    }
    return classif;
}

double VariationalBayesianGaussianMixtureClassifier::predict(const vec& instance) {
    double l0 = hmog.estimatePosteriorLogLikelihood(0,instance);
    double l1 = hmog.estimatePosteriorLogLikelihood(1,instance);
    double p0, p1;
    if ((l0>700.0) || (l1>700.0) || (l0<-700.0) || (l1<-700.0)) {
        if (l0>l1) p0 = 1.0;
        else p0 = 0.0;
        p1 = 1.0 - p0;
    } else {
        p0 = trunc_exp(l0);
        p1 = trunc_exp(l1);
    }
    p0 /= p0 + p1;
    return 1.0 - p0 * 2.0;
}


