/***************************************************************************
 *   Authors:                                                              *
 *   - Nicolas Brodu                                                       *
 *                                                                         *
 *   Copyrights:                                                           *
 *   - Nicolas Brodu (initial base code, some modifications in SVN)        *
 *   - INRIA (modifications in SVN)                                        *
 *                                                                         *
 *   This file is free software; you can redistribute it and/or modify     *
 *   it under the terms of the GNU General Public License as               *
 *   published by the Free Software Foundation; either version 2 of the    *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU Lesser General Public License for more details.                   *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 ***************************************************************************/

#include <iostream>
#include <algorithm>
#include <fstream>
#include <vector>
#include <set>
#include <map>
using namespace std;
#include <math.h>

#include <itpp/itbase.h>
using namespace itpp;

// Code structure is one subdirectory per classifier type, main interface in the main dir
// Each subdirectory is as independent as possible so could be reused in an external application
#include "LinearMinSumSquaredError/LinearMinSumSquaredError.h"
#include "LinearFisherDiscriminant/LinearFisherDiscriminant.h"
#include "HiddenLayerPerceptron/HiddenLayerPerceptron.h"
#include "MaximumLikelihoodGaussianMixture/MaximumLikelihoodGaussianMixture.h"
#include "VariationalBayesianGaussianMixture/VariationalBayesianGaussianMixture.h"
#include "NearestNeighbours/NearestNeighbours.h"
#include "SupportVectorMachines/SupportVectorMachines.h"
// TODO: kernel fisher discriminant ?

#include "PredictionCombiners.h"
#include "AdaBoost.h"

#include "Utility.h"


// Initialized in main with all known factories
vector<ClassifierFactory*> classifierFactories;

void help(const char* progname) {
cerr<<
"This program can train a variety of classifiers and apply them to a test data set.\n"
"\n"
"Usage: " << progname << " [options] train_data train_classes test_data test_classes output_trainclassif output_testclassif\n"
"\n"
"The argument files are in plain text format with space delimiters:\n"
"\n"
"  For the data files, there are num_data*num_feature*feature_size values on each line: data1_feature1 data1_feature2 data1_feature3... data1_featureN data2_feature1 data2_feature2 ... Each feature may have its own size. Ex: Feature1 might be a 3-dimensional vector, Feature2 a scalar... And each feature has its own data units and scale. See options -f, -e. By default, only one feature (of deduced feature_size) is used per data instance, num_data is deduced from the class file.\n"
"\n"
"  For the data files, each row represents a different instance in time. If your data is static (only one time instance), then just write only one row in the data files\n"
"\n"
"  For the classes files, there must be num_data values, which are fixed for all times. Use either 1 or -1 for the two classes. Sorry, more classes are not available for now. If you need to train using more classes, then you may always apply one-against-others or one-against-one classification and aggregate the results.\n"
"\n"
"Note that a classifier is trained for every time sample. The training and test data files number of rows must be the same. Results are also aggregated through time in order to get a final classification value.\n"
"\n"
"If you don't know the test classes, then just write any random class in the test_classes file. You'll still get a final predicted classification, just ignore the error made by this classification.\n"
"\n"
"\n"
"0. Option Syntax:\n"
"\n"
"Options indicated by (G) act globally and follow their respective simple syntax.\n"
"\n"
"Options indicated by (S) also act globally if using their respective basic syntax, so don't bother with the rest of this section if you only wish to use global options.\n"
"\n"
"(S) options might also be specialized for selected feature:method pairs. Assume -X<int,int,int> is an option taking three numbers as argument. If you just write -X 1,2,3 then X applies globally, for all features and prediction methods. However you might also use the following:\n"
"-X feature:method:opt1,opt2,opt3;feature:method:opt4,opt5,opt6;...\n"
"In that case the options will only apply to the given features and methods. See below for available methods. In addition:\n"
". feature might be omitted. In that case the given arguments apply to all features though only for the selected method. Example: -X method:opt1,opt2,opt3\n"
". method might be omitted. In that case the given arguments apply to all feature-space methods (like -1, see below), though only for the selected feature. Example: -X feature::opt1,opt2,opt3. Note: method might also be -1 or -2 explicitly.\n"
". If both are omitted, then the global syntax is recovered and the option acts globally.\n"
"Note: Only methods/features that are explicitly written (that is, not implicit by omission or -1/-2) are automatically instanciated, see the -m/-c options. The reason for this is that -q for example, applied globally, does not need to run all classifiers just to make them silent...\n"
"\n"
"\n"
"1. Informative options:\n"
"\n"
"  -v: (S) Verbose mode. Prints more information related to the algorithms parameters and execution.\n"
"\n"
"  -q: (S) Quiet mode (takes precedence over verbose). Prints only the final performances and other critical information.\n"
"\n"
"\n"
"2. Data related options:\n"
"\n"
"  -f <int,int...>: (G) Successive features take the given number of columns. Ex: -f 3,4 specifies that each instance has two features, respectively taking 3 and 4 columns of the data file. Classifiers are trained on each feature separately. You may then want to combine the classifiers (see below)."
"\n"
"  -e: (G) Exclude some features globally. Read the data file description above and the -f option. You may then specify features to exclude, counting from 1. Ex: -f3,4,2 -e2 specifies three features taking 3, 4 and 2 columns respectively, and then removes feature 2 that takes 4 columns.\n"
"\n"
"  -t <int>: (G) Start from time instance t, counting from 1 (default is 1, take all instances).\n"
"\n"
"  -w <int>: (G) Sliding window size (default=1). Past data values are cumulated over the window size in order to get more instances for the training.\n"
"\n"
"  -s <low_bound,high_bound>. (S) Set the data scale. The training data is by default scaled in the -1 1 range before applying the classification method. The same scaling factor is then applied on the test data (which of course, may make it go outside the -1 1 range). A range 0,0 disables scaling, so -s0,0 disables scaling globally. Use the specialization syntax to set scaling only for some feature:method combinations, and just remember -1,1 is the default, not 0,0. Note: The source code could possibly handle scaling ranges for every dimension of every feature, it's just not available on the command line. A future extension might consider parameter files for that.\n"
"\n"
"\n"
"3. Classification methods:\n"
"\n"
"  -m <[feature:]method,...>: (G) Methods to use. Several methods may be specified, optionally restricted to some features. Available methods are selected by their number, see below. Method specific options are distinguished by capital letters and can be given separately on the command-line. Method specific options can be specialized (S).\n"
"\n"
"Methods assuming data are homogeneous feature vectors in some space\n";
for (unsigned int m=0; m<classifierFactories.size(); ++m) if (classifierFactories[m]->getType()==ClassifierFactory::FeatureVector)
classifierFactories[m]->printClassifierHelpText(cerr);
cerr <<
"    -1: Apply all feature vector methods, one after another. You might find the -q option useful in this case.\n"
"\n"
"Methods assuming data are samples from the same probability distribution\n";
for (unsigned int m=0; m<classifierFactories.size(); ++m) if (classifierFactories[m]->getType()==ClassifierFactory::DistributionSample)
classifierFactories[m]->printClassifierHelpText(cerr);
cerr <<
"    -2: Apply all probability distribution sample methods, one after another.\n"
"\n"
"Note: -m is a global option: unlike the (S) syntax, all feature/method valid combinations specified here will be run (precisely including -1/-2 aliases for multiple methods). For this reason ',' are used to separate arguments, not ';'. See also the -c option.\n"
"\n"
"Note: Linear Fisher Discriminant Analysis and min sum of squares project to the same space in the case of +1/-1 classes. Both implementations are kept: the FDA allows for different thresholds for the decision, and it provides a class separability value for each time step which is used for combining decisions over time, while the sum-of-square classifier just aggregates using the maximum distance.\n"
"\n"
"\n"
"4. Cross-validation:\n"
"\n"
"  -x <int>: (S) Use v-fold cross-validation to estimate the best parameters for the chosen methods. This option does nothing if the method has no parameter (ex: linear min sum of squares). The given number is v, the number of subsets that are considered for the cross-validation. Use v=number of training instances to use leave-one-out cross-validation. Default: Do not use cross-validation.\n"
"\n"
"All method specific options marked by (*) can be automatically tuned by cross-validation. If a specific value is given while cross-validation is selected, then that value is kept constant and cross-validation is performed on the remaining parameters. This might be especially useful to save time.\n"
"When three values are given instead of one this specifies a range and number of steps for the cross-validation. Ex: -m6 -x10 -K3,10,8 will use the K-nearest neighbours method, with a 10-fold cross-validation to search for the best number of neighbours in the integer range K = 3..10, bounds included.\n"
"The -C and -G options divide the search range using a logarithmic scale. Ex: -C 1e-1,1e2,4 applies the cross-validation successively for C=0.1, C=1, C=10, and C=100. A finer search is then conducted nearby the optimal value found by the first rough pass, using the same number of steps. See the libsvm documentation.\n"
"Default search ranges are used during cross-validation if the (*) options are not specified.\n"
"\n"
"\n"
"5. Combining classifiers\n"
"\n"
"  -b <iter[,resample,[force]]>: (S) Adaptive boosting (AdaBoost) is applied for each feature:method pair that is given by the (S) syntax, for the given number of iterations. The individual feature:method results might additionally be combined by -c to get a final result if desired.\n"
"\n"
"Note: Adaboost works by weighting the training instances according to their difficulty of being learned. Some classification methods support weigthed instances (these are indicated by (w)), but others do not. In this case Adaboost simulates weighting by resampling the training set using the normalized weights as a probability distribution for the resampling. The optional 'resample' option gives the size of the resampled training set. The default is the same size as the original training set. Resampling is normally only used as a fallback when the target classification method does not support weighting. The 'force' optional argument might be set to 1 in order to force resampling, whether the boosted classifier supports weighting or not. This might for example be used to change the apparent size of the training set, or to compare the results between the resampling and the direct weighting schemes.\n"
"\n"
"  -c <int,feature:method,...>: (G) Combine the results of the various predictions using one of the following combination rules. The first value indicates the combination rule. Then pairs of feature:method indicates which prediction results are combined (see example below). Available combination rules are:\n";
PredictionCombinerFactory::printCombinerHelpText(cout);
cerr <<
"Note: -c is a global (G) option since combination precisely comes after the specialized computations, so the combination rule cannot be an argument of a specialized method. Hence the use of , and not ; for the parameters. Methods specified here are considered automatically and need not (but might) be specified in the -m option list.\n"
"\n"
"  -a <int>: (S). Aggregates the classification results through time. The integer argument is a combination rule, applied to the predictions at different time steps for each individual feature:method. See the -c option above for a description of available rules.\n"
"\n"
"Note: If both -c and -a are given, additional combinations are available:\n"
". The time aggregation of each feature:method is combined by -c, at each time step. This is done automatically, and using the -c rule.\n"
". The result of the -c combination at each time step may be aggregated though time. In order to specify (S) a rule for this combination, use 0:0 as the feature:method specification in the -a option or use a global (G) syntax rule (i.e. -1 feature). If no rule is given either explicitly or implicitly this combination is not computed.\n"
". If both previous combinations are computed, they are combined again using the rule specified by -c.\n"
"\n"
<<flush;
}

// TODO: for each classifier help text, say what confidence is used.

struct OneSpecializedSyntaxToken {
    int feature;
    int method;
    string option;
};
typedef vector<OneSpecializedSyntaxToken> SpecializedSyntaxTokens;

SpecializedSyntaxTokens parseOptionSpecializedSyntax(string optarg) {

    SpecializedSyntaxTokens ret;

    // two-level parser -- first split tokens on (S) syntax separator ;
    for (unsigned int i=0; i<optarg.size(); ++i) if (optarg[i]==';') optarg[i]=' ';
    stringstream tokenizer(optarg);
    string Sspec;
    while (tokenizer >> Sspec) {
        vector<int> colidx;
        for (unsigned int i=0; i<Sspec.size(); ++i) if (Sspec[i]==':') colidx.push_back(i);
        if (colidx.size()>2) {
            cerr << "Invalid syntax: "<<Sspec<<endl;
            exit(0);
        }
        OneSpecializedSyntaxToken token;
        // no feature / no method => global
        if (colidx.size()==0) {
            token.feature = -1;
            token.method = -1;
            token.option = Sspec;
        }
        // no feature
        else if (colidx.size()==1 || colidx[0]==0) {
            for (unsigned int i=0; i<colidx.size(); ++i) Sspec[colidx[i]] = ' ';
            stringstream tokenizer2(Sspec);
            token.feature = -1;
            tokenizer2 >> token.method;
            tokenizer2 >> token.option;
        }
        // no method only when consecutive ::
        else if (colidx.size()==2 && colidx[1]==colidx[0]+1) {
            for (unsigned int i=0; i<colidx.size(); ++i) Sspec[colidx[i]] = ' ';
            stringstream tokenizer2(Sspec);
            tokenizer2 >> token.feature;
            token.method = -1;
            tokenizer2 >> token.option;
        // full case
        } else {
            for (unsigned int i=0; i<colidx.size(); ++i) Sspec[colidx[i]] = ' ';
            stringstream tokenizer2(Sspec);
            tokenizer2 >> token.feature;
            tokenizer2 >> token.method;
            tokenizer2 >> token.option;
        }
        ret.push_back( token );
    }

    return ret;
}

// One data set per feature per time
// each_time< each_feature< each_instance < ndim_data > > >
// vector< vector< vector< vec > > > trainingSet;
struct InstanceSet {
    vector< vec > instances; // each instance is n-dimensional
    // formula is rescaled = (original - dataShift) * scalingFactors + scaledShift;
    // but scaled range is specific for each feature:method
    // => precompute data min/max so translation is easy later on
    vec dataMin; // of size ndim
    vec dataRange; // of size ndim
};

struct AdaBoostParameters {
    int niter;
    int nresample;
    int forceResample;
    AdaBoostParameters() : niter(0), nresample(0), forceResample(0) {}
};

struct FeatureMethodSpecifics {
    ClassifierInterface* classifier;
    // Data scale: from -1 to 1 by default
    vec low_bound;
    vec scaled_range;
    PredictionCombiner* trainAggregateThroughTime; // time aggregation rule. 0 = disabled
    PredictionCombiner* testAggregateThroughTime; // time aggregation rule. 0 = disabled
    bool isCombined;
    FeatureMethodSpecifics() : classifier(0), trainAggregateThroughTime(0), testAggregateThroughTime(0), isCombined(false) {low_bound = -1.0; scaled_range = 2.0;}
    FeatureMethodSpecifics(ClassifierInterface* c) : classifier(c), trainAggregateThroughTime(0), testAggregateThroughTime(0), isCombined(false) { low_bound = -1.0; scaled_range = 2.0;}
};

int main(int argc, char** argv) {

    std::ios::sync_with_stdio(false);

    // Hardcoded known classifiers. Add your own here.
    classifierFactories.push_back(new LinearMinSumSquaredErrorFactory());
    classifierFactories.push_back(new LinearFisherDiscriminantFactory());
    classifierFactories.push_back(new HiddenLayerPerceptronFactory());
    classifierFactories.push_back(new MaximumLikelihoodGaussianMixtureFactory());
    classifierFactories.push_back(new VariationalBayesianGaussianMixtureFactory());
    classifierFactories.push_back(new NearestNeighboursFactory(NearestNeighboursFactory::Euclidian));
    classifierFactories.push_back(new SupportVectorMachinesFactory(libsvm::LINEAR));
    classifierFactories.push_back(new SupportVectorMachinesFactory(libsvm::RBF));
    classifierFactories.push_back(new SupportVectorMachinesFactory(libsvm::POLY));
    classifierFactories.push_back(new SupportVectorMachinesFactory(libsvm::SIGMOID));
    // methods acting on probability distribution samples
    classifierFactories.push_back(new NearestNeighboursFactory(NearestNeighboursFactory::KolmogorovSmirnov));

    vector< pair<int, string> > remainingOptions;

    int startTime = 0;
    int slidingWindow = 1;

    // -1 => auto-set
    int nfeatures = -1;
    vector<int> featuresDim;
    vector<int> isExcludedFeature;

    int c; opterr = 0;
    string optionsString ="v::q::f:e:s:t:w:m:x:a:b:c:r:";
    for (unsigned int m=0; m<classifierFactories.size(); ++m) {
        optionsString += classifierFactories[m]->getClassifierOptions();
    }

    int globalRandomSeed = 42;

    while ((c=getopt(argc,argv,optionsString.c_str()))!=-1) switch(c) {

// data options only for now, before reading the files
        case 'f': {
            char *c = optarg;
            while (*c) {if (*c==',') *c= ' '; ++c;}
            stringstream tokenizer(optarg);
            int fdim;
            while (tokenizer >> fdim) {
                if (fdim<=0) {
                    cerr << "Invalid feature dimension: " << fdim << endl;
                    help(argv[0]);
                    return 1;
                }
                featuresDim.push_back(fdim);
            }
            nfeatures = featuresDim.size();
            break;
        }

        case 'e': {
            char *c = optarg;
            while (*c) {if (*c==',') *c= ' '; ++c;}
            stringstream tokenizer(optarg);
            int feature; // counted from 1
            while (tokenizer >> feature) {
                if (feature<=0) {
                    cerr << "Invalid excluded feature number: " << feature << endl;
                    help(argv[0]);
                    return 1;
                }
                // final size will be set below
                if (isExcludedFeature.size() < (unsigned int)feature) isExcludedFeature.resize(feature, 0);
                isExcludedFeature[feature-1] = 1;
            }
            break;
        }

        case 'r': {
            globalRandomSeed = atoi(optarg);
            if (globalRandomSeed<1) {
                cerr << "Invalid random seed: " << optarg << endl;
                help(argv[0]);
                return 1;
            }
            break;
        }

        case 't': startTime = atoi(optarg) - 1; break;

        case 'w': {
            slidingWindow = atoi(optarg);
            if (slidingWindow<1) {
                cerr << "Invalid sliding window size: " << optarg << endl;
                help(argv[0]);
                return 1;
            }
            break;
        }

        case '?': {
            // not recognized by getopt => help
            help(argv[0]);
            return 1;
        }

        // Store for later use, once methods are instanciated
        default: {
            remainingOptions.push_back( make_pair(c, optarg?string(optarg):string()) );
            break;
        }

    }

    Utility::RandomInit(globalRandomSeed);

    // fill up exclusion list if needed
    if (nfeatures>0) {
        if (isExcludedFeature.size() < (unsigned int)nfeatures) isExcludedFeature.resize(nfeatures, 0);
        if (isExcludedFeature.size() > (unsigned int)nfeatures) {
            cerr << "Invalid excluded feature number: " << isExcludedFeature.size() << endl;
            help(argv[0]);
            return 1;
        }
    }

    if (argc - optind < 6) {
        help(argv[0]);
        return 1;
    }

    ifstream train_data(argv[optind]);
    if (!train_data) {
        cerr << "Cannot read train_data file: " << argv[optind] << endl;
        return 1;
    }

    ifstream train_classes(argv[optind+1]);
    if (!train_classes) {
        cerr << "Cannot read train_classes file: " << argv[optind+1] << endl;
        return 1;
    }

    ifstream test_data(argv[optind+2]);
    if (!test_data) {
        cerr << "Cannot read test_data file: " << argv[optind+2] << endl;
        return 1;
    }

    ifstream test_classes(argv[optind+3]);
    if (!test_classes) {
        cerr << "Cannot read test_classes file: " << argv[optind+3] << endl;
        return 1;
    }

    ofstream output_trainclassif(argv[optind+4]);
    if (!output_trainclassif) {
        cerr << "Cannot write to output_trainclassif file: " << argv[optind+4] << endl;
        return 1;
    }

    ofstream output_testclassif(argv[optind+5]);
    if (!output_testclassif) {
        cerr << "Cannot write to output_testclassif file: " << argv[optind+5] << endl;
        return 1;
    }

    typedef vector< InstanceSet > FeatureSet;      // indices are feature numbers
    typedef vector< FeatureSet > TemporalDataSet;  // indices are time steps

    TemporalDataSet trainSet, testSet;  // Could introduce pair<> to force them to have the same size, but let's keep it simple
    ivec trainClasses, testClasses;      // of size respective number of instances

    // duplicated info... but handy
    int ntraininstances, ntestinstances;

    // scoped variable, deallocated when unused => convert to itpp objects
    {

        string line; double value;

        vector<int> cppclasses;
        while (train_classes && !train_classes.eof()) {
            getline(train_classes, line);
            if (line.empty()) continue;
            stringstream linereader(line);
            int classnum;
            while (linereader >> classnum) cppclasses.push_back(classnum);
            if ((cppclasses.back() != -1) && (cppclasses.back() != 1)) {cerr << argv[optind+1] << ": Invalid class number, should be -1 or 1!" << endl; return 1;}
        }
        train_classes.close();

        ntraininstances = cppclasses.size();
        if (!ntraininstances) {cerr << argv[optind+1] << ": No data!" << endl; return 1;}

        trainClasses.set_size(ntraininstances);
        for (int row = 0; row<ntraininstances; ++row) trainClasses(row) = cppclasses[row];

        cppclasses.clear();
        while (test_classes && !test_classes.eof()) {
            getline(test_classes, line);
            if (line.empty()) continue;
            stringstream linereader(line);
            int classnum;
            while (linereader >> classnum) cppclasses.push_back(classnum);
            if ((cppclasses.back() != -1) && (cppclasses.back() != 1)) {cerr << argv[optind+3] << ": Invalid class number, should be -1 or 1!" << endl; return 1;}
        }
        test_classes.close();

        ntestinstances = cppclasses.size();
        if (!ntestinstances) {cerr << argv[optind+3] << ": No data!" << endl; return 1;}

        testClasses.set_size(ntestinstances);
        for (int row = 0; row<ntestinstances; ++row) testClasses(row) = cppclasses[row];

        int sumofdimfeatures = -1;
        int ndataperline = -1;

        int linenum = 0; int timestep = 0;
        while (train_data && !train_data.eof()) {
            getline(train_data, line);
            ++linenum;
            if (line.empty()) continue;
            ++timestep;
            stringstream linereader(line);
            vector<double> values;
            while (linereader >> value) values.push_back(value);
            if (ndataperline == -1) {
                ndataperline = values.size();

                // initialise number of features from the first line: first constraint: sum of dimensions should be an integer multiple
                sumofdimfeatures = ndataperline / ntraininstances;
                if ((ntraininstances * sumofdimfeatures) != ndataperline ) {
                    cerr << argv[optind] << ": Number of values is not a multiple of number of data instances (as read from class file)!" << endl;
                    return 1;
                }
                // not specified by options => only one feature per instance, taking all dims
                if (nfeatures==-1) {
                    nfeatures = 1;
                    featuresDim.resize(1);
                    featuresDim[0] = sumofdimfeatures;
                    isExcludedFeature.resize(1);
                    isExcludedFeature[0] = 0;
                } else {
                    // check sumofdimfeatures is really that
                    int realsumofdf = 0;
                    for (int f=0; f<nfeatures; ++f) realsumofdf += featuresDim[f];
                    if (realsumofdf != sumofdimfeatures) {
                        cerr << argv[optind] << ": Cannot split values amongst specified feature dimensions!" << endl;
                        return 1;
                    }
                }

            }
            if ((unsigned int)(ndataperline) != values.size()) {
                cerr << argv[optind] << ", line " << linenum<< ": Inconsistent training data file! All rows must have the same number of values." << endl;
                return 1;
            }

            // ignore data before startTime;
            if (timestep<startTime) continue;

            // use resize and not push_back to avoid useless copy later on
            trainSet.resize(trainSet.size()+1);
            FeatureSet& featureSet = trainSet.back();
            featureSet.resize(nfeatures);

            // Store data in final structures
            int fidxShift = 0;
            for (int f=0; f<nfeatures; ++f) {
                if (isExcludedFeature[f]) {
                    fidxShift += featuresDim[f];
                    continue;
                }
                InstanceSet& iset = featureSet[f];
                iset.instances.resize(ntraininstances);
                // set size
                iset.dataMin.set_size(featuresDim[f]);
                iset.dataRange.set_size(featuresDim[f]);
                for (int i=0; i<ntraininstances; ++i) {
                    iset.instances[i].set_size(featuresDim[f]);
                    for (int j=0; j<featuresDim[f]; ++j) {
                        iset.instances[i](j) = values[i*sumofdimfeatures + fidxShift + j];
                        // data min/max (use dataRange for max temporarily)
                        if ((i==0) || (iset.instances[i](j) < iset.dataMin(j))) iset.dataMin(j) = iset.instances[i](j);
                        if ((i==0) || (iset.instances[i](j) > iset.dataRange(j))) iset.dataRange(j) = iset.instances[i](j);
                    }
                }
                // Fix range from max-min
                for (int j=0; j<featuresDim[f]; ++j) iset.dataRange(j) -= iset.dataMin(j);
                fidxShift += featuresDim[f];
            }
        }
        train_data.close();

        if (trainSet.empty()) {
            cerr << argv[optind] << ": No data!" << endl;
            return 1;
        }

        // test set must be consistent
        ndataperline = sumofdimfeatures * ntestinstances;
        linenum = 0; timestep = 0;
        while (test_data && !test_data.eof()) {
            getline(test_data, line);
            ++linenum;
            if (line.empty()) continue;
            ++timestep;
            if (timestep<startTime) continue;
            // Generalize through timestep: timestep-startTime is index in trainSet
            if (timestep-startTime >= (int)trainSet.size()) timestep = startTime + trainSet.size()-1;
            stringstream linereader(line);
            vector<double> values;
            while (linereader >> value) values.push_back(value);
            if ((unsigned int)(ndataperline) != values.size()) {
                cerr << argv[optind+2] << ", line " << linenum<< ": Inconsistent test data file! All rows must have the correct number of values, fixed by the number of features and the number of test instances." << endl;
                return 1;
            }

            // use resize and not push_back to avoid useless copy later on
            testSet.resize(testSet.size()+1);
            FeatureSet& featureSet = testSet.back();
            featureSet.resize(nfeatures);

            // Store data in final structures
            int fidxShift = 0;
            for (int f=0; f<nfeatures; ++f) {
                if (isExcludedFeature[f]) {
                    fidxShift += featuresDim[f];
                    continue;
                }
                InstanceSet& iset = featureSet[f];
                iset.instances.resize(ntestinstances);
                for (int i=0; i<ntestinstances; ++i) {
                    iset.instances[i].set_size(featuresDim[f]);
                    for (int j=0; j<featuresDim[f]; ++j) iset.instances[i](j) = values[i*sumofdimfeatures + fidxShift + j];
                }
                // apply the same scaling factors as in the training data set
                InstanceSet& tiset = trainSet[timestep-startTime][f];
                iset.dataMin = tiset.dataMin;
                iset.dataRange = tiset.dataRange;
                fidxShift += featuresDim[f];
            }
        }
        test_data.close();

        if (testSet.empty()) {
            cerr << argv[optind+2] << ": No test data!" << endl;
            return 1;
        }
    }


    // Now we know definite number of features
    // => complete method specification with -m, -c, -b and (S) syntax
    // but do nothing at this point, just fill up all methods.
    // The goal is to ensure uniqueness in instanciation later on.
    // Only then options are re-parsed and can apply on concrete instances.

    // temporary holder while the option list has not been completely run through.
    // map: index_in_feature_set => map of (indices_in_factory_vector, adaboost iterations)
    map<int, map<int,AdaBoostParameters> > methodSpec;

    // feature_index: method IDs for that feature (warning mixed conventions)
    map<int, set<int> > combiSpec;
    // TODO: use set
    vector<PredictionCombiner*> trainCombiners;
    vector<PredictionCombiner*> testCombiners;
    // combine all rules, then combine again through time
    PredictionCombiner* trainCombinedAggregator = 0;
    PredictionCombiner* testCombinedAggregator = 0;

    // Complete method initialization from option list
    for (unsigned int oidx = 0; oidx < remainingOptions.size(); ++oidx) {
        // string copy
        string optarg = remainingOptions[oidx].second;
        switch(remainingOptions[oidx].first) {

// Direct method selection -- applies to all given features
            case 'm': {
                for (unsigned int i=0; i<optarg.size(); ++i) if (optarg[i]==',') optarg[i]=' ';
                stringstream tokenizer(optarg);
                string fmspec;
                while (tokenizer >> fmspec) {
                    vector<int> colidx;
                    for (unsigned int i=0; i<fmspec.size(); ++i) if (fmspec[i]==':') {colidx.push_back(i); fmspec[i]=' ';}
                    if (colidx.size()>1) {
                        cerr << "Invalid -m syntax: " << remainingOptions[oidx].second << endl;
                        help(argv[0]);
                        return 1;
                    }
                    stringstream tokenizer2(fmspec);
                    int featureNumber, methodNumber;
                    // no feature => all features
                    if (colidx.empty()) featureNumber = -1;
                    else tokenizer2 >> featureNumber;
                    tokenizer2 >> methodNumber;

                    bool handled = false;
                    for (unsigned int m=0; m<classifierFactories.size(); ++m) {
                        if (methodNumber==classifierFactories[m]->getMethodId()
                        || (methodNumber==-1 && classifierFactories[m]->getType()==ClassifierFactory::FeatureVector)
                        || (methodNumber==-2 && classifierFactories[m]->getType()==ClassifierFactory::DistributionSample)
                        ) {
                            // accumulate method specs over all options
                            // using a set ensures uniqueness of instanciation later on
                            // Adaboost is set below, initialize to 0 here
                            if (featureNumber==-1) for (int f=0; f<nfeatures; ++f) methodSpec[f][m] = AdaBoostParameters();
                            else methodSpec[featureNumber-1][m] = AdaBoostParameters();
                            handled = true;
                        }
                    }
                    if (!handled) {
                        cerr << "Invalid method number: " << methodNumber << endl;
                        help(argv[0]);
                        return 1;
                    }
                }
                break;
            }

// combination is just parsed, so long as we don't have real instances
// But it forces instanciation

            case 'c': {
                if (!trainCombiners.empty()) {
// TODO: use sets for trainCombiners and testCombiners
// TODO: allow specialisation of combiSpec and methodSpec for each rule. Here only -1 is supported
                    cerr << "Sorry, multiple explicit combination rules are not supported\nPlease use the -1 combination rule for now (combinations have negligible costs so performing them all is not that awful)\nAlso check for a newer version!" << endl;
                    return 0;
                }
                for (unsigned int i=0; i<optarg.size(); ++i) if (optarg[i]==',') optarg[i]=' ';
                stringstream tokenizer(optarg);
                // parse global combination rule now, this information does not depend on feature/method
                int combinationRule;
                tokenizer >> combinationRule;
                if (combinationRule==-1) {
                    for (int i=1; i<=PredictionCombinerFactory::numRules; ++i) {
                        trainCombiners.push_back(PredictionCombinerFactory::instanciateCombiner(i, ntraininstances));
                        testCombiners.push_back(PredictionCombinerFactory::instanciateCombiner(i, ntestinstances));
                    }
                } else {
                    if (combinationRule<=0 || combinationRule >PredictionCombinerFactory::numRules) {
                        cerr << "Invalid combination rule ID (-c option): " << optarg << endl;
                        help(argv[0]);
                        return 1;
                    }
                    trainCombiners.push_back(PredictionCombinerFactory::instanciateCombiner(combinationRule, ntraininstances));
                    testCombiners.push_back(PredictionCombinerFactory::instanciateCombiner(combinationRule, ntestinstances));
                }
                // This part is common to all rules, but we checked that only -1 is valid for multiple rules => OK
                string combiSpecArg;
                while (tokenizer >> combiSpecArg) {
                    for (unsigned int i=0; i<combiSpecArg.size(); ++i) if (combiSpecArg[i]==':') combiSpecArg[i]=' ';
                    stringstream tokenizer2(combiSpecArg);
                    int featureNumber, methodNumber;
                    tokenizer2 >> featureNumber;
                    tokenizer2 >> methodNumber;
                    bool handled = false;
                    if ((featureNumber==-1) || (featureNumber>0)) for (unsigned int m=0; m<classifierFactories.size(); ++m) {
                        if (methodNumber==classifierFactories[m]->getMethodId()
                        || (methodNumber==-1 && classifierFactories[m]->getType()==ClassifierFactory::FeatureVector)
                        || (methodNumber==-2 && classifierFactories[m]->getType()==ClassifierFactory::DistributionSample)
                        ) {
                            if (featureNumber==-1) for (int f=0; f<nfeatures; ++f) {
                                methodSpec[f][m] = AdaBoostParameters();
                                combiSpec[f].insert(classifierFactories[m]->getMethodId());
                            }
                            else {
                                methodSpec[featureNumber-1][m] = AdaBoostParameters();
                                combiSpec[featureNumber-1].insert(classifierFactories[m]->getMethodId());
                            }
                            handled = true;
                        }
                    }
                    if (!handled) {
                        cerr << "Invalid combination specification: " << remainingOptions[oidx].second << endl;
                        help(argv[0]);
                        return 1;
                    }
                }
                break;
            }

            // Necessarily one of the recognized options at this point, amongst "vqs:x:a:b:"
            // or an option provided by a classifier method, all of which support the (S) syntax
            default: {
                SpecializedSyntaxTokens tokens = parseOptionSpecializedSyntax(optarg);
                // Only explicit instanciation at this point, -1 and -2 disabled
                for (unsigned int tok=0; tok<tokens.size(); ++tok) if (tokens[tok].feature>0) for (unsigned int m=0; m<classifierFactories.size(); ++m) {
                    if (tokens[tok].method==classifierFactories[m]->getMethodId()) {
                        methodSpec[tokens[tok].feature-1][m] = AdaBoostParameters();
                    }
                }
            }

        }
    }

    // OK, so now we can ensure uniqueness in instanciation
    // apply adaboot on all given methods that are marked to be instanciated but before instanciation
    for (unsigned int oidx = 0; oidx < remainingOptions.size(); ++oidx) {
        if (remainingOptions[oidx].first != 'b') continue;
        SpecializedSyntaxTokens tokens = parseOptionSpecializedSyntax(remainingOptions[oidx].second);
        for (unsigned int tok=0; tok<tokens.size(); ++tok) {
            stringstream tokenizer(tokens[tok].option);
            int adaboost_iter = 0, adaboost_resample = 0, adaboost_force = 0;
            tokenizer >> adaboost_iter;
            tokenizer >> adaboost_resample;
            tokenizer >> adaboost_force;
            if (adaboost_iter<=1) {
                cerr << "There must be more than one iteration for Adaptive Boosting to work: " << optarg << endl;
                help(argv[0]);
                return 1;
            }
            // resample and force params are optional, might be null (handled by the AdaBoost class)
            for (unsigned int m=0; m<classifierFactories.size(); ++m) {
                if (tokens[tok].method==classifierFactories[m]->getMethodId()
                || (tokens[tok].method==-1 && classifierFactories[m]->getType()==ClassifierFactory::FeatureVector)
                || (tokens[tok].method==-2 && classifierFactories[m]->getType()==ClassifierFactory::DistributionSample)
                ) {
                    // method match => set number of adaboost iterations
                    // careful: only applies if the entry exists (-1, -2...), do not force instanciation
                    if (tokens[tok].feature==-1) for (int f=0; f<nfeatures; ++f) {
                        map<int,AdaBoostParameters>::iterator mit = methodSpec[f].find(m);
                        if (mit!=methodSpec[f].end()) {
                            mit->second.niter = adaboost_iter;
                            mit->second.nresample = adaboost_resample;
                            mit->second.forceResample = adaboost_force;
                        }
                    }
                    else {
                        map<int,AdaBoostParameters>::iterator mit = methodSpec[tokens[tok].feature-1].find(m);
                        if (mit!=methodSpec[tokens[tok].feature-1].end()) {
                            mit->second.niter = adaboost_iter;
                            mit->second.nresample = adaboost_resample;
                            mit->second.forceResample = adaboost_force;
                        }
                    }
                }
            }
        }
    }

    // Process at least one metod?
    if (methodSpec.empty()) {
        cerr << "Please specify at least one classification method to use!" << endl;
        help(argv[0]);
        return 1;
    }

    // For each handled feature, the methods to use
    // map: index_in_features_array => (method ID => classifier_instances)
    map<int, map<int, FeatureMethodSpecifics> > methods;

    // Can now instanciate all methods, possibly with adaboost wrapper
    for (map<int, map<int,AdaBoostParameters> >::iterator fit = methodSpec.begin(); fit!=methodSpec.end(); ++fit) {
        int fidx = fit->first;
        //if (featureNumber<0) {cerr << "Internal error: no negative feature index should be present at this point. Please report a bug." << endl; return 1;}
        // run through method spec
        for (map<int,AdaBoostParameters>::iterator mit = fit->second.begin(); mit != fit->second.end(); ++mit) {
            ClassifierFactory* factory = classifierFactories[mit->first];
            FeatureMethodSpecifics& fms = methods[fidx][factory->getMethodId()];
            fms.low_bound.set_size(featuresDim[fidx]);
            fms.scaled_range.set_size(featuresDim[fidx]);
            fms.low_bound = -1.0;
            fms.scaled_range = 2.0;
            // avoid side effects from the random sequence: each classifier behaves as if run alone
            Utility::RandomInit(globalRandomSeed);
            // Adaboost ?
            if (mit->second.niter>0)
                // Yes, provide the factory and number of iterations (number of "weak classifier" instances)
                fms.classifier = new AdaBoost(factory, mit->second.niter, mit->second.nresample, mit->second.forceResample);
            else
                // Directly instanciate the classifier
                fms.classifier = factory->instanciateClassifier();
        }
    }



    for (map<int, set<int> >::iterator cit = combiSpec.begin(); cit!=combiSpec.end(); ++cit) {
        for (set<int>::iterator midit = cit->second.begin(); midit != cit->second.end(); ++midit) {
            // Instance is ensured to exist at this point
            methods[cit->first][*midit].isCombined = true;
        }
    }

    // helpers... OK, admittedly that's not pretty.
    vector<int> methodIDsForMinus1;
    vector<int> methodIDsForMinus2;
    for (unsigned int m=0; m<classifierFactories.size(); ++m) {
        if (classifierFactories[m]->getType()==ClassifierFactory::FeatureVector)
            methodIDsForMinus1.push_back(classifierFactories[m]->getMethodId());
        if (classifierFactories[m]->getType()==ClassifierFactory::DistributionSample)
            methodIDsForMinus2.push_back(classifierFactories[m]->getMethodId());
    }

    // global variables in addition to method specific verbosity
    bool quiet = false;
    bool verbose = false;

    // Complete method initialization from option list
    for (unsigned int oidx = 0; oidx < remainingOptions.size(); ++oidx) {
        // string copy
        string optarg = remainingOptions[oidx].second;
        switch(remainingOptions[oidx].first) {

            // -m, -c, -b have been processed already. Could also be removed above, but without changing the vector order
            case 'm':
            case 'c':
            case 'b':
                break;

            // "vqs:x:" and method options => all using (S) syntax
            default: {
                // set global v/q flags if needed
                if (remainingOptions[oidx].first=='v') verbose = true;
                if (remainingOptions[oidx].first=='q') quiet = true;
                SpecializedSyntaxTokens tokens = parseOptionSpecializedSyntax(optarg);
                for (unsigned int tok=0; tok<tokens.size(); ++tok) {
                    OneSpecializedSyntaxToken& token = tokens[tok];
                    // Special case for the -a option: virtual feature (0), or implicit feature (-1)
                    if (tokens[tok].feature<=0 && remainingOptions[oidx].first=='a' && nfeatures>1) {
                        stringstream tokenizer(token.option);
                        int combinedAggregationRule;
                        tokenizer >> combinedAggregationRule;
                        PredictionCombiner* combi = PredictionCombinerFactory::instanciateCombiner(combinedAggregationRule, ntraininstances);
                        if (!combi) {
                            cerr << "Invalid combination rule (-a option): " << combinedAggregationRule << endl;
                            help(argv[0]);
                            return 1;
                        }
                        if (trainCombinedAggregator) delete trainCombinedAggregator;
                        trainCombinedAggregator = combi;
                        if (testCombinedAggregator) delete testCombinedAggregator;
                        testCombinedAggregator = PredictionCombinerFactory::instanciateCombiner(combinedAggregationRule, ntestinstances);
                        // fall through only if -1 to broadcast to all real features
                        if (tokens[tok].feature!=-1) continue;
                    }
                    vector<int> features;
                    if (tokens[tok].feature==-1) for (int f=0; f<nfeatures; ++f) features.push_back(f);
                    else features.push_back(tokens[tok].feature);
                    vector<int> methodIDs;
                    if (tokens[tok].method==-1) for (unsigned int m=0; m<methodIDsForMinus1.size(); ++m) methodIDs.push_back(methodIDsForMinus1[m]);
                    else if (tokens[tok].method==-2) for (unsigned int m=0; m<methodIDsForMinus2.size(); ++m) methodIDs.push_back(methodIDsForMinus2[m]);
                    else methodIDs.push_back(tokens[tok].method);
                    for (vector<int>::iterator fit = features.begin(); fit != features.end(); ++fit) {
                        // only applies to existing instances at this point
                        map<int, map<int, FeatureMethodSpecifics> >::iterator fsearchit = methods.find(*fit);
                        if (fsearchit == methods.end()) continue;
                        for (vector<int>::iterator mit = methodIDs.begin(); mit != methodIDs.end(); ++mit) {
                            map<int, FeatureMethodSpecifics>::iterator msearchit = fsearchit->second.find(*mit);
                            if (msearchit == fsearchit->second.end()) continue;
                            // Found an existing instance for that combination
                            ClassifierInterface* classifier = msearchit->second.classifier;
                            switch (remainingOptions[oidx].first) {
                                case 'v': classifier->setVerbose(true); break;
                                case 'q': classifier->setQuiet(true); break;
                                case 'x': {
                                    stringstream tokenizer(token.option);
                                    int vfoldnum;
                                    tokenizer >> vfoldnum;
                                    if ((vfoldnum<=0) || (vfoldnum>ntraininstances)) {
                                        cerr << "Invalid number of v-fold cross-validation numbers: " << token.option << endl;
                                        help(argv[0]);
                                        return 1;
                                    }
                                    classifier->setVFoldCrossValidation(vfoldnum);
                                    break;
                                }
                                case 's': {
                                    for (unsigned int i=0; i<token.option.size(); ++i) if (token.option[i]==',') token.option[i]=' ';
                                    stringstream tokenizer(token.option);
                                    double lb, hb;
                                    tokenizer >> lb;
                                    tokenizer >> hb;
                                    if (hb<lb) {
                                        cerr << "Invalid scaling range: " << lb << "," << hb << endl;
                                        help(argv[0]);
                                        return 1;
                                    }
                                    msearchit->second.low_bound = lb;
                                    msearchit->second.scaled_range = hb - lb;
                                    break;
                                }
                                case 'a': {
                                    stringstream tokenizer(token.option);
                                    int aggregationRule;
                                    tokenizer >> aggregationRule;
                                    PredictionCombiner* combi = PredictionCombinerFactory::instanciateCombiner(aggregationRule, ntraininstances);
                                    if (!combi) {
                                        cerr << "Invalid aggregation rule (-a option): " << aggregationRule << endl;
                                        help(argv[0]);
                                        return 1;
                                    }
                                    if (msearchit->second.trainAggregateThroughTime) delete msearchit->second.trainAggregateThroughTime;
                                    msearchit->second.trainAggregateThroughTime = combi;
                                    if (msearchit->second.testAggregateThroughTime) delete msearchit->second.testAggregateThroughTime;
                                    msearchit->second.testAggregateThroughTime = PredictionCombinerFactory::instanciateCombiner(aggregationRule, ntestinstances);
                                    break;
                                }
                                // one of the options handled by the method
                                default: classifier->parseOption(remainingOptions[oidx].first,token.option);
                            }
                        }
                    }
                }
                break;
            }
        }
    }


    if (quiet) verbose = false; // quiet takes precedence

    for (map<int, map<int, FeatureMethodSpecifics> >::iterator fit = methods.begin(); fit != methods.end(); ++fit) {
        for (map<int, FeatureMethodSpecifics>::iterator mit = fit->second.begin(); mit != fit->second.end(); ++mit) {
            if (mit->second.classifier->getQuiet()) mit->second.classifier->setVerbose(false);
        }
    }

    if (verbose) {
        cout << "Files loaded successfully." << endl;
        cout << ntraininstances << " training instances, " << ntestinstances << " test instances." << endl;
        cout << nfeatures << " features per instance, with dimension" << (nfeatures>1?"s":"") << ":";
        bool hasExclusion = false;
        for (int f=0; f<nfeatures; ++f) {
            cout << " ";
            if (isExcludedFeature[f]) {
                hasExclusion = true;
                cout << "E";
            }
            cout << featuresDim[f];
        }
        if (hasExclusion) cout << " (E = excluded)";
        cout << endl;
        cout << trainSet.size() << " time samples for the train set, " << testSet.size() << " for the test set." << endl;
        cout << endl;
        cout << "Starting from time " << startTime << ", using a sliding window of size " << slidingWindow << " time units." << endl;
    }

    vec classcount(2); classcount.zeros();
    for (int i=0; i<ntraininstances; ++i) ++classcount(((int)(trainClasses(i))+1)/2);


    for (unsigned int timestep = 0; timestep<testSet.size(); ++timestep) {

        // Train over the sliding window (possibly of size 1)
        int trainIdxStop = (timestep<trainSet.size()) ? timestep : (trainSet.size()-1);
        int trainIdxStart = trainIdxStop - slidingWindow + 1;
        if (trainIdxStart<0) trainIdxStart = 0;

        if (!quiet) cout << "Processing time step " << (timestep+1) << endl;

        // combine classifiers across rules: reset each time step
        // see below for aggregation through time
        for (unsigned int i=0; i<trainCombiners.size(); ++i) trainCombiners[i]->reset();
        for (unsigned int i=0; i<testCombiners.size(); ++i) testCombiners[i]->reset();

/*
        cout << "Scaling ranges written as feature:min:max";
        for (int f=0; f<nfeatures; ++f) cout << " " << (f+1) << ":" << scaledLowBound[f](0) << ":" << scaledHighBound[f](0);
*/

        // For each SELECTED feature...
        for (map<int, map<int, FeatureMethodSpecifics> >::iterator fit = methods.begin(); fit != methods.end(); ++fit) {
            // aliases
            int feature = fit->first;
            int featureDim = featuresDim[feature];

            if (!quiet) cout << "Processing feature: " << (feature+1) << endl;

            // For each SELECTED method for that feature...
            for (map<int, FeatureMethodSpecifics>::iterator mit = fit->second.begin(); mit != fit->second.end(); ++mit) {

                // Avoid interference from the random sequence, so each feature:method produces exactly the same result as if run alone
                // Use multiple runs with different seeds to study the influence of the random sequence
                Utility::RandomInit(globalRandomSeed);

                // aliases
                ClassifierInterface* classifier = mit->second.classifier;
                const vec& scaled_range = mit->second.scaled_range;
                const vec& low_bound = mit->second.low_bound;

                if (!quiet) cout << "Using classification method: " << classifier->getName() << endl;

                bool isScaled = mit->second.scaled_range(0)>0.0;
                // TODO: either support per-dimension settings, or replace range vectors by scalars
                if (verbose) {
                    if (!isScaled) cout << "Data is not scaled." << endl;
                    else cout << "Scaling range: " << mit->second.low_bound(0) << " to " << (mit->second.low_bound(0) + mit->second.scaled_range(0)) << endl;
                }

                int ntotaltraininstances = (trainIdxStop - trainIdxStart + 1) * ntraininstances;
                classifier->prepareForTraining(ntotaltraininstances, featureDim);
                // alias for consistency
                int ntotaltestinstances = ntestinstances;

                for (int tridx = trainIdxStart; tridx <= trainIdxStop; ++tridx) {
                    InstanceSet& instanceSet = trainSet[tridx][feature];
                    for (int i=0; i<ntraininstances; ++i) {
                        if (isScaled) {
                            vec scaledData(featureDim);
                            // scaled = (data - dmin) * (high_bound - low_bound) / (dmax - dmin) + low_bound;
                            // scaled = (data - dmin) * scaled_range / dataRange + low_bound;
                            for (int j=0; j<featureDim; ++j) {
                                if (instanceSet.dataRange(j)>0.0) {
                                    scaledData(j) = (instanceSet.instances[i](j) - instanceSet.dataMin(j)) * scaled_range(j) / instanceSet.dataRange(j) + low_bound(j);
                                }
                            }
                            classifier->train(scaledData, trainClasses(i));
                        }
                        else classifier->train(instanceSet.instances[i], trainClasses(i));
                    }
                }

                // finishTraining also performs cross-validation if necessary, before predictions
                double classifierConfidence = classifier->finishTraining();

                // Now compute error on the train set
                int numCorrectClassification = 0;
                vec prediction(ntraininstances);
                for (int tridx = trainIdxStart; tridx <= trainIdxStop; ++tridx) {
                    InstanceSet& instanceSet = trainSet[tridx][feature];
                    for (int i=0; i<ntraininstances; ++i) {
                        if (isScaled) {
                            vec scaledData(featureDim);
                            for (int j=0; j<featureDim; ++j) {
                                if (instanceSet.dataRange(j)>0.0) {
                                    scaledData(j) = (instanceSet.instances[i](j) - instanceSet.dataMin(j)) * scaled_range(j) / instanceSet.dataRange(j) + low_bound(j);
                                }
                            }
                            prediction(i) = classifier->predict(scaledData);
                        }
                        else prediction(i) = classifier->predict(instanceSet.instances[i]);

                        numCorrectClassification += (prediction(i) * trainClasses(i) > 0.0); // same sign
                    }
                }

                double perf = numCorrectClassification * 100.0 / ntotaltraininstances;
                if (!quiet) cout << "Train classification performance: " << perf << "%" << endl;
                if (fit != methods.begin() || mit != fit->second.begin()) output_trainclassif << " "; output_trainclassif << perf;

                // combine this rule with others (across rules) at each time step (not across time)
                if (mit->second.isCombined) for (unsigned int i=0; i<trainCombiners.size(); ++i) trainCombiners[i]->addPrediction(prediction, classifierConfidence);

                // aggregate the result of this rule through time.
                if (mit->second.trainAggregateThroughTime) {
                    mit->second.trainAggregateThroughTime->addPrediction(prediction, classifierConfidence);

                    // Compute aggregated error on the train set so far
                    vec combinePrediction = mit->second.trainAggregateThroughTime->combinePredictions();
                    int numCorrectClassification = 0;
                    for (int i=0; i<ntraininstances; ++i) {
                        numCorrectClassification += (combinePrediction(i) * trainClasses(i) > 0.0); // same sign
                    }
                    perf = numCorrectClassification * 100.0 / ntotaltraininstances;
                    if (!quiet) cout << "Aggregated train classification performance: " << perf << "%" << endl;
                    output_trainclassif << " " << perf;
                }


                // Now compute error on the test set
                numCorrectClassification = 0;
                prediction.set_size(ntestinstances);
                {
                    InstanceSet& instanceSet = testSet[timestep][feature];
                    for (int i=0; i<ntestinstances; ++i) {
                        if (isScaled) {
                            vec scaledData(featureDim);
                            for (int j=0; j<featureDim; ++j) {
                                if (instanceSet.dataRange(j)>0.0) {
                                    scaledData(j) = (instanceSet.instances[i](j) - instanceSet.dataMin(j)) * scaled_range(j) / instanceSet.dataRange(j) + low_bound(j);
                                }
                            }
                            prediction(i) = classifier->predict(scaledData);
                        }
                        else prediction(i) = classifier->predict(instanceSet.instances[i]);

                        numCorrectClassification += (prediction(i) * testClasses(i) > 0.0); // same sign
                    }
                }

                perf = numCorrectClassification * 100.0 / ntotaltestinstances;
                if (!quiet) cout << "Test classification performance: " << perf << "%" << endl;
                if (fit != methods.begin() || mit != fit->second.begin()) output_testclassif << " "; output_testclassif << perf;

                // combine this rule with others (across rules) at each time step (not across time)
                if (mit->second.isCombined) for (unsigned int i=0; i<testCombiners.size(); ++i) testCombiners[i]->addPrediction(prediction, classifierConfidence);

                // aggregate the result of this rule through time.
                if (mit->second.testAggregateThroughTime) {
                    mit->second.testAggregateThroughTime->addPrediction(prediction, classifierConfidence);

                    // Compute aggregated error on the test set so far
                    vec combinePrediction = mit->second.testAggregateThroughTime->combinePredictions();
                    int numCorrectClassification = 0;
                    for (int i=0; i<ntestinstances; ++i) {
                        numCorrectClassification += (combinePrediction(i) * testClasses(i) > 0.0); // same sign
                    }
                    perf = numCorrectClassification * 100.0 / ntotaltestinstances;
                    if (!quiet) cout << "Aggregated test classification performance: " << perf << "%" << endl;
                    output_testclassif << " " << perf;
                }


            }
        }

        // combined rules prediction performance
        for (unsigned int ci=0; ci<trainCombiners.size(); ++ci) {

            // Compute combined error on the train set
            vec combinePrediction = trainCombiners[ci]->combinePredictions();

            // aggregate combined predictions through time if required
            if (trainCombinedAggregator) {
                // TODO: only the first rule is used for now...
                if (trainCombiners.size() > 1) {
                    if (!quiet) cout << "Warning: Only aggregating the result of the first combination rule through time. This is a limitation of the current version." << endl;
                }
                if (ci==0) trainCombinedAggregator->addPrediction(combinePrediction, 1.0);
            }

            int numCorrectClassification = 0;
            for (int i=0; i<ntraininstances; ++i) {
                numCorrectClassification += (combinePrediction(i) * trainClasses(i) > 0.0); // same sign
            }
            double perf = numCorrectClassification * 100.0 / ntraininstances;
            if (!quiet) cout << "Combination rule " << (ci+1) << " train classification performance: " << perf << "%" << endl;
            output_trainclassif << " " << perf;
        }
        for (unsigned int ci=0; ci<testCombiners.size(); ++ci) {

            // Compute combined error on the train set
            vec combinePrediction = testCombiners[ci]->combinePredictions();

            // aggregate combined predictions through time if required
            if (testCombinedAggregator) {
                // TODO: only the first rule is used for now...
                if (testCombiners.size() > 1) {
                    if (!quiet) cout << "Warning: Only aggregating the result of the first combination rule through time. This is a limitation of the current version." << endl;
                }
                if (ci==0) testCombinedAggregator->addPrediction(combinePrediction, 1.0);
            }

            int numCorrectClassification = 0;
            for (int i=0; i<ntestinstances; ++i) {
                numCorrectClassification += (combinePrediction(i) * testClasses(i) > 0.0); // same sign
            }
            double perf = numCorrectClassification * 100.0 / ntestinstances;
            if (!quiet) cout << "Combination rule " << (ci+1) << " test classification performance: " << perf << "%" << endl;
            output_testclassif << " " << perf;
        }


/*      // TODO: update for multiple combination rules support

        // Double-combo check
        bool hasCombinationOfAggregate = false;
        if (trainCombiner) {
            trainCombiner->reset();
            testCombiner->reset();
            for (map<int, map<int, FeatureMethodSpecifics> >::iterator fit = methods.begin(); fit != methods.end(); ++fit) {
                for (map<int, FeatureMethodSpecifics>::iterator mit = fit->second.begin(); mit != fit->second.end(); ++mit) {
                    if (mit->second.isCombined && mit->second.trainAggregateThroughTime) {
                        trainCombiner->addPrediction(mit->second.trainAggregateThroughTime->combinePredictions(), 1.0);
                        testCombiner->addPrediction(mit->second.testAggregateThroughTime->combinePredictions(), 1.0);
                        hasCombinationOfAggregate = true;
                    }
                }
            }
            if (hasCombinationOfAggregate) {
                vec combinePrediction = trainCombiner->combinePredictions();
                int numCorrectClassification = 0;
                for (int i=0; i<ntraininstances; ++i) {
                    numCorrectClassification += (combinePrediction(i) * trainClasses(i) > 0.0); // same sign
                }
                double perf = numCorrectClassification * 100.0 / ntraininstances;
                if (!quiet) cout << "Combination of time-aggregated methods train classification performance: " << perf << "%" << endl;
                output_trainclassif << " " << perf;
                // prepare for final double-combo, if any
                trainCombiner->reset();
                trainCombiner->addPrediction(combinePrediction, 1.0);
                combinePrediction = testCombiner->combinePredictions();
                numCorrectClassification = 0;
                for (int i=0; i<ntestinstances; ++i) {
                    numCorrectClassification += (combinePrediction(i) * testClasses(i) > 0.0); // same sign
                }
                perf = numCorrectClassification * 100.0 / ntestinstances;
                if (!quiet) cout << "Combination of time-aggregated methods test classification performance: " << perf << "%" << endl;
                output_testclassif << " " << perf;
                testCombiner->reset();
                testCombiner->addPrediction(combinePrediction, 1.0);
            }
        }

        // Aggregation of combined rules prediction performance
        if (trainCombinedAggregator) {
            vec combinePrediction = trainCombinedAggregator->combinePredictions();
            int numCorrectClassification = 0;
            for (int i=0; i<ntraininstances; ++i) {
                numCorrectClassification += (combinePrediction(i) * trainClasses(i) > 0.0); // same sign
            }
            double perf = numCorrectClassification * 100.0 / ntraininstances;
            if (!quiet) cout << "Aggregation of combined methods train classification performance: " << perf << "%" << endl;
            output_trainclassif << " " << perf;
            if (hasCombinationOfAggregate) trainCombiner->addPrediction(combinePrediction, 1.0);
        }
        if (testCombinedAggregator) {
            vec combinePrediction = testCombinedAggregator->combinePredictions();
            int numCorrectClassification = 0;
            for (int i=0; i<ntestinstances; ++i) {
                numCorrectClassification += (combinePrediction(i) * testClasses(i) > 0.0); // same sign
            }
            double perf = numCorrectClassification * 100.0 / ntestinstances;
            if (!quiet) cout << "Aggregation of combined methods test classification performance: " << perf << "%" << endl;
            output_testclassif << " " << perf;
            if (hasCombinationOfAggregate) testCombiner->addPrediction(combinePrediction, 1.0);
        }

        // Double-combo check
        if (hasCombinationOfAggregate && trainCombinedAggregator) {
            vec combinePrediction = trainCombiner->combinePredictions();
            int numCorrectClassification = 0;
            for (int i=0; i<ntraininstances; ++i) {
                numCorrectClassification += (combinePrediction(i) * trainClasses(i) > 0.0); // same sign
            }
            double perf = numCorrectClassification * 100.0 / ntraininstances;
            if (!quiet) cout << "Final mix of {aggregation of combinations, combinations of aggregations} train performance: " << perf << "%" << endl;
            output_trainclassif << " " << perf;
            combinePrediction = testCombiner->combinePredictions();
            numCorrectClassification = 0;
            for (int i=0; i<ntestinstances; ++i) {
                numCorrectClassification += (combinePrediction(i) * testClasses(i) > 0.0); // same sign
            }
            perf = numCorrectClassification * 100.0 / ntestinstances;
            if (!quiet) cout << "Final mix of {aggregation of combinations, combinations of aggregations} test performance: " << perf << "%" << endl;
            output_testclassif << " " << perf;
        }
*/
        output_trainclassif << endl;
        output_testclassif << endl;
    }

    return 0;
}

