/***************************************************************************
 *   Authors:                                                              *
 *   - Nicolas Brodu                                                       *
 *                                                                         *
 *   Copyrights:                                                           *
 *   - Nicolas Brodu (initial author of incremfa)                          *
 *   - INRIA (modifications/diffs in SVN)                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation; either version 2 of the    *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU Lesser General Public License for more details.                   *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef CYCLIC_ARRAY_H
#define CYCLIC_ARRAY_H

#include <string.h>

/// Fifo cyclic buffer
/// All inlined for efficiency
template<class T> struct CyclicArrayTemplate {

    T* array;

    /// Include member variables in the data array => better data cache hit ratio!
    /// Assumes sizeof(T) >= sizeof(int) for this to work
    /// If this is not the case, define the following constant
//#define CYCLIC_ARRAY_MEMBER_VARIABLES

#ifdef CYCLIC_ARRAY_MEMBER_VARIABLES

    int _firstidx, _capacity, _size;

    inline int& firstidx() {
        return _firstidx;
    }
    inline int& capacity() {
        return _capacity;
    }
    inline int& size() {
        return _size;
    }

#define CYCLIC_ARRAY_SHIFT 0
#else

    inline int& firstidx() {
        return *reinterpret_cast<int*>(array-3);
    }
    inline int& capacity() {
        return *reinterpret_cast<int*>(array-2);
    }
    inline int& size() {
        return *reinterpret_cast<int*>(array-1);
    }

#define CYCLIC_ARRAY_SHIFT 3
#endif

    CyclicArrayTemplate(int cap) {
        array = new T[cap+CYCLIC_ARRAY_SHIFT];
        array+=CYCLIC_ARRAY_SHIFT;
        firstidx() = 0;
        this->capacity() = cap;
        size() = 0;
    }

    ~CyclicArrayTemplate() {
        delete[] (array - CYCLIC_ARRAY_SHIFT);
    }

    inline bool isFull() {
        return size()==capacity();
    }
    
    inline void zeroFill() {
        memset(array, 0, sizeof(T)*capacity());
        firstidx() = 0;
        size() = capacity();
    }

    /// Swap this array with another one without making data copy (exchange pointers)
    inline void swap(CyclicArrayTemplate& ca) {
        T* tmparray = array;
        array = ca.array;
        ca.array = tmparray;
#ifdef CYCLIC_ARRAY_MEMBER_VARIABLES
        int tmp_firstidx = firstidx();
        int tmp_capacity = capacity();
        int tmp_size = size();
        firstidx() = ca.firstidx();
        capacity() = ca.capacity();
        size() = ca.size();
        ca.firstidx() = tmp_firstidx;
        ca.capacity() = tmp_capacity;
        ca.size() = tmp_size;
#endif
    }

    /// return direct access to linear array if possible for the given range
    /// other wise, return 0
    inline T* linear(int idx, int sz) {
        idx += firstidx();
        if (idx+sz<=capacity()) return array+idx;
        if (idx>=capacity()) idx-=capacity();
        if (idx+sz<=capacity()) return array+idx;
        return 0;
    }

    /// ensure the array points to a contiguous memory zone
    inline void linearize() {
        if (firstidx() == 0) return;
        T* linarray = new T[capacity()+CYCLIC_ARRAY_SHIFT];
        linarray+=CYCLIC_ARRAY_SHIFT;
        memcpy(linarray, array + firstidx(), sizeof(T) * (capacity() - firstidx()));
        memcpy(linarray + capacity() - firstidx(), array, sizeof(T) * firstidx());
#if CYCLIC_ARRAY_SHIFT > 0
        memcpy(linarray - CYCLIC_ARRAY_SHIFT, array - CYCLIC_ARRAY_SHIFT, sizeof(T)*CYCLIC_ARRAY_SHIFT);
#endif
        delete[] (array - CYCLIC_ARRAY_SHIFT);
        array = linarray;
        firstidx() = 0;
    }

    inline T& operator[](int idx) {
        // No bound check (idx >? _size, <? 0 )
        // if ((idx<0) || (idx>=_size)) throw "Array out of bounds!";
        idx += firstidx();
#ifdef NO_CMOV_ASM
        if (idx>=capacity()) idx-=capacity();
#else
        register int dupreg = idx;
        asm volatile (
//        "movl %[idx], %[dupreg]\n\t"     // tmp = idx
        "subl %[cap], %[idx]\n\t"        // idx -= capacity
        "cmovl %[dupreg], %[idx]\n\t"    // if (idx<0) idx = tmp
        : [idx] "+r" (idx), [dupreg] "+r" (dupreg)
        : [cap] "g" (capacity())
        :
        );
#endif
        return array[idx];
    }

    /// Warning: returned reference is only valid till push_back erases it (cyclic buffer).
    ///          If it is valid to begin with... (no internal error checking for size() > 0)
    inline T& pop_front() {
        // no bound check size() >= 1
        --size();
        T& ret = array[firstidx()++];
        if (firstidx()>=capacity()) firstidx() = 0;
        return ret;
    }

    inline void push_back(const T& value) {
        (*this)[size()++] = value;
    }

    /// Does both push_back & pop_front in one pass
    /// Works only when the buffer is full
    inline void push_cyclic(const T& value) {
        array[firstidx()] = value;
        if (++firstidx()>=capacity()) firstidx() = 0;
    }
};

/// Default type to double
typedef CyclicArrayTemplate<double> CyclicArray;

#endif
