/***************************************************************************
 *   Authors:                                                              *
 *   - Nicolas Brodu                                                       *
 *                                                                         *
 *   Copyrights:                                                           *
 *   - INRIA                                                               *
 *   - Nicolas Brodu (some personal additions, see svn log)                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation; either version 2 of the    *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU Lesser General Public License for more details.                   *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <math.h>

#include "DensityEstimator.h"

// Include Vikas C. Raykar fast kernel density estimation routine. Thanks Vikas :)
#include "UnivariateDensityDerivative.h"
// Include C implementation of netlib minpack lmfit function adapted from the sourceforge project of the same name
#include "lmfit.h"

using namespace std;

// =======================
// the proxies / interface
// =======================

double estimateDensity(const std::vector<double>& samples, const std::vector<double>& x, std::vector<double>& density, double kernelWidth, double epsilon) {
    return estimateDensity(&samples[0], samples.size(), &x[0], x.size(), &density[0], kernelWidth, epsilon);
}

double estimateDensity(const double* samples, int Ns, const double* x, int Nx, double* density, double kernelWidth, double epsilon) {
    if (Ns<1) return kernelWidth;
    double smin = samples[0];
    double smax = smin;

    for (int i=1; i<Ns; ++i) {
        if (samples[i] > smax) smax = samples[i];
        if (samples[i] < smin) smin = samples[i];
    }

    // rescale all data in the 0..1 range
    vector<double> scaledSamples(Ns);

    double scaleFactor = 1.0;
    if (smax>smin) scaleFactor = 1.0 / (smax - smin);

    for (int i=0; i<Ns; ++i) {
        scaledSamples[i] = (samples[i] - smin) * scaleFactor;
        // floating point numbers are not mathematical reals...
        if (scaledSamples[i]<0.0) scaledSamples[i] = 0.0;
        if (scaledSamples[i]>1.0) scaledSamples[i] = 1.0;
    }

    vector<double> scaledX(Nx);
    for (int i=0; i<Nx; ++i) scaledX[i] = (x[i] - smin) * scaleFactor;

    double kw = estimateDensityScaledData(&scaledSamples[0], Ns, &scaledX[0], Nx, &density[0], kernelWidth*scaleFactor, epsilon);

    for (int i=0; i<Nx; ++i) density[i] *= scaleFactor;

    return kw/scaleFactor;
}

double computeOptimalKernelWidth(const std::vector<double>& samples, double epsilon) {
    return computeOptimalKernelWidth(&samples[0], samples.size(), epsilon);
}

double computeOptimalKernelWidth(const double* samples, int Ns, double epsilon) {
    if (Ns<1) return 0.0; // to be determined later on when data is available
    double smin = samples[0];
    double smax = smin;

    for (int i=1; i<Ns; ++i) {
        if (samples[i] > smax) smax = samples[i];
        if (samples[i] < smin) smin = samples[i];
    }

    // rescale all data in the 0..1 range
    vector<double> scaledSamples(Ns);

    double scaleFactor = 1.0;
    if (smax>smin) scaleFactor = 1.0 / (smax - smin);

    for (int i=0; i<Ns; ++i) {
        scaledSamples[i] = (samples[i] - smin) * scaleFactor;
        // floating point numbers are not mathematical reals...
        if (scaledSamples[i]<0.0) scaledSamples[i] = 0.0;
        if (scaledSamples[i]>1.0) scaledSamples[i] = 1.0;
    }

    return computeOptimalKernelWidthScaledData(samples, Ns, epsilon) / scaleFactor;
}

double estimateDensityScaledData(const std::vector<double>& samples, const std::vector<double>& x, std::vector<double>& density, double kernelWidth, double epsilon) {
    return estimateDensityScaledData(&samples[0], samples.size(), &x[0], x.size(), &density[0], kernelWidth, epsilon);
}

double computeOptimalKernelWidthScaledData(const std::vector<double>& samples, double epsilon) {
    return computeOptimalKernelWidthScaledData(&samples[0], samples.size(), epsilon);
}



// ============
// Wrapper code
// ============

struct lmfit_data_for_evaluate {
    double constant1;
    double constant2;
    int N;
    double epsilon;
    double* X;
};

// Code inspired from the matlab file fast_h_function.m by Vikas C. Raykar, LGPL v2.1+
// Adapted to the lmfit world.
static void lmfit_evaluate(double* h, int, double* fvec, void* data_ptr, int*) {

    lmfit_data_for_evaluate* data = (lmfit_data_for_evaluate*)data_ptr;

    // code translated from matlab
    double lambda = data->constant2 * pow(*h, 0.7142857142857142857); // 5/7
    if (isnan(lambda) || isinf(lambda)) {
        *fvec = 0;
        return;
    }
    vector<double> D4(data->N);
    UnivariateDensityDerivative udd4(data->N,data->N,data->X,data->X,lambda,4,data->epsilon,&D4[0]);
    udd4.Evaluate();
    double phi4 = 0.0;
    for (int i=0; i<data->N; ++i) phi4 += D4[i];
    phi4 /= data->N - 1.0;
    *fvec = *h - data->constant1 * pow(phi4, -0.2);

}

// May be used for debug, if necessary
static void lmfit_printout ( int n_par, double* par, int m_dat, double* fvec, 
                       void *data, int iflag, int iter, int nfev )
{
}


// =================
// The real function
// =================

double computeOptimalKernelWidthScaledData(const double* samples, int Ns, double epsilon) {

    // compute the standard deviation of the scaled data
    double mean = 0.0;
    double var = 0.0;
    for (int i=0; i<Ns; ++i) {
        mean += samples[i];
        var += samples[i] * samples[i];
    }
    mean /= Ns;
    var = (var - mean*mean*Ns) / (Ns+1.0);
    if (var<=0.0) {
        // var = 0.0; // Null variance leads to null sigma, -inf phi6/8, and bad kernel.
        return 0.0;
    }

    double sigma = sqrt(var);

    // Estimate the density functionals Phi_6 and Phi_8 using the normal scale rule.
    const double c1 = -0.528927734576021519; // -15/(16*sqrt(pi))
    const double c2 = 1.851247071016075317; // 105/(32*sqrt(pi))
    double phi6 = c1*pow(sigma,-7.0);
    double phi8 = c2*pow(sigma,-9.0);

    // Estimate the density functionals Phi_4 and Phi_6 using the kernel density
    // estimators with the optimal bandwidth based on the asymptotic MSE.
    const double c3 = -2.393653682408596068; // -6/sqrt(2*pi)
    const double c4 = 11.96826841204298034;  // 30/sqrt(2*pi)

    double g1=pow(c3/(phi6*Ns),0.1428571428571428571); // 1/7
    double g2=pow(c4/(phi8*Ns),0.1111111111111111111); // 1/9

    if (isnan(g1) || isnan(g2)) return 0.0;
    if (isinf(g1) || isinf(g2)) return 0.0;

    vector<double> D4(Ns);
    UnivariateDensityDerivative udd4(Ns,Ns,const_cast<double*>(samples),const_cast<double*>(samples),g1,4,epsilon,&D4[0]);
    udd4.Evaluate();
    vector<double> D6(Ns);
    UnivariateDensityDerivative udd6(Ns,Ns,const_cast<double*>(samples),const_cast<double*>(samples),g2,6,epsilon,&D6[0]);
    udd6.Evaluate();
    double phi4 = 0.0;
    phi6 = 0.0;
    for (int i=0; i<Ns; ++i) {
        phi4 += D4[i];
        phi6 += D6[i];
    }
    phi4 /= (Ns-1.0);
    phi6 /= (Ns-1.0);

    lmfit_data_for_evaluate data;

    data.constant1 = pow(0.2820947917738781435 / Ns, 0.2); // (1/(2*sqrt(pi)*Ns))^(1/5);
    data.constant2 = pow(-8.485281374238570293 * phi4 / phi6, 0.1428571428571428571);  // (-6*sqrt(2)*phi4/phi6)^(1/7);
    data.N = Ns;
    data.epsilon = epsilon;
    data.X = const_cast<double*>(samples);

    if (isnan(data.constant1) || isnan(data.constant2)) return 0.0;
    if (isinf(data.constant1) || isinf(data.constant2)) return 0.0;

    // Initial guess of the bandwidth
    double bandwidth = data.constant1 * pow(phi4, -0.2);

    if (isnan(bandwidth) || isinf(bandwidth)) return 0.0;

    // output variables
    double fvec, diag, fjac, qtf, wa1,wa2,wa3,wa4;
    int info,nfev,ipvt;

    // calling the optimizer
    lm_lmdif(       // see parameters documentation in lmfit.cpp
        1,          // m        number of functions
        1,          // n        number of variables
        &bandwidth, // x        target variable to optimize from initial guess
        &fvec,      // fvec     value of the objective function at x
        1e-14,      // ftol     tolerance on the error on the function evaluation
        1e-14,      // xtol     tolerance on the error on x
        1e-14,      // gtol     tolerance on the error on the jacobian
        100,        // maxfev   max number of iterations
        1e-14,      // epsfcn   step length for forward differencing
        &diag,      // diag     unused here due to mode=1
        1,          // mode     1 = scale variables internally.
        100.,       // factor   bound on step size, recommended value
        &info,      // info     will hold diagnostics on the result
        &nfev,      // nfev     will be set to the number of function evaluations that were used
        &fjac,      // fjac     storage space
        &ipvt,      // ipvt     storage space
        &qtf,       // qtf      storage space
        &wa1,       // wa1      storage space
        &wa2,       // wa2      storage space
        &wa3,       // wa3      storage space
        &wa4,       // wa4      storage space
        &lmfit_evaluate,    // function that computes current error on x
        &lmfit_printout,    // function that can be used to print progress for debug
        &data
    );

    if (isnan(bandwidth) || isinf(bandwidth)) return 0.0;

    return bandwidth;
}

double estimateDensityScaledData(const double* samples, int Ns, const double* x, int Nx, double* density, double kernelWidth, double epsilon) {
    if (kernelWidth<=0.0) kernelWidth = computeOptimalKernelWidthScaledData(samples, Ns, epsilon);
    UnivariateDensityDerivative udd(Ns,Nx,const_cast<double*>(samples),const_cast<double*>(x),kernelWidth,0,epsilon,density);
    udd.Evaluate();
    return kernelWidth;
}
