/***************************************************************************
 *   Authors:                                                              *
 *   - Nicolas Brodu                                                       *
 *                                                                         *
 *   Copyrights:                                                           *
 *   - INRIA                                                               *
 *   - Nicolas Brodu (some personal additions, see svn log)                *
 *                                                                         *
 *   This block of code is independent from the rest of the project        *
 *   - DensityEstimation.*                                                 *
 *   - lmfit.*                                                             *
 *   - UnivariateDensityDerivative.*                                       *
 *   See the individual files for comments                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation; either version 2 of the    *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU Lesser General Public License for more details.                   *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef DENSITY_ESTIMATOR_H
#define DENSITY_ESTIMATOR_H

#include <vector>

/// This function estimates the density at each of the given points using AMISE-optimal kernel density estimation up to the desired precision
/// @param samples Observed data samples -- INPUT parameter
/// @param x The points at which the density must be estimated -- INPUT parameter
/// @param density The density values at the given points -- OUTPUT parameter of the same size as x.
/// @param kernelWidth The size of the kernel. The default value 0 means compute an optimal value automatically, but this is relatively computationally expensive. Alternatively, you might wish to call computeOptimalKernelWidth yourself, or provide another fixed kernel width if you're not satisfied with either the default or the computation time. Be careful not to under/oversmooth however...
/// @param epsilon Precision desired on the kernel width
/// @return the kernel width that was used, automatically determined if the argument was null
double estimateDensity(const std::vector<double>& samples, const std::vector<double>& x, std::vector<double>& density, double kernelWidth = 0.0, double epsilon = 1e-6);

/// Idem, using a plain pointer/size API. See comments above
double estimateDensity(const double* samples, int Ns, const double* x, int Nx, double* density, double kernelWidth = 0.0, double epsilon = 1e-6);


/// Compute the AMISE-optimal kernel width for estimating the density of the given points
/// This routine relies on Vikas C. Raykar library for "Fast optimal bandwidth selection for kernel density estimation".
/// Vikas C. Raykar and Ramani Duraiswami, In Proceedings of the sixth SIAM International Conference on Data Mining, Bethesda, April 2006, pp. 524-528. This gives a fast kernel derivative estimate that is exact at an arbitrary required precision.
/// Then the Shether Jones Solve-the-equation plug-in method is applied on the kernel derivative in order to get the kernel width. "A reliable data-based bandwidth selection method for kernel density estimation", S.J. Sheather and M.C. Jones, Journal of the Royal Statistics Society B, 53:683-690, 1991
/// The last step of the plug-in method requires solving a non-linear least squares problem.
/// The Levenberg-Marquardt algorithm from lmfit / netlib / minpack is thus used to get the AMISE solution.
/// @param samples The data samples for which an optimal kernel width is desired
/// @param epsilon Precision required on the solution. Default is 1e-6
double computeOptimalKernelWidth(const std::vector<double>& samples, double epsilon = 1e-6);

/// Idem, using a plain pointer/size API
double computeOptimalKernelWidth(const double* samples, int Ns, double epsilon = 1e-6);


/// Assumes the x data has already been scaled in the 0..1 range. Use this method ONLY if this is the case. See above comments.
double estimateDensityScaledData(const std::vector<double>& samples, const std::vector<double>& x, std::vector<double>& density, double kernelWidth = 0.0, double epsilon = 1e-6);
double estimateDensityScaledData(const double* samples, int Ns, const double* x, int Nx, double* density, double kernelWidth = 0.0, double epsilon = 1e-6);

/// Assumes the x data has already been scaled in the 0..1 range. Use this method ONLY if this is the case. See above comments.
double computeOptimalKernelWidthScaledData(const std::vector<double>& samples, double epsilon = 1e-6);
double computeOptimalKernelWidthScaledData(const double* samples, int Ns, double epsilon = 1e-6);


#endif
