/***************************************************************************
 *   Authors:                                                              *
 *   - Nicolas Brodu                                                       *
 *                                                                         *
 *   Copyrights:                                                           *
 *   - Nicolas Brodu (initial author of incremfa)                          *
 *   - INRIA (modifications/diffs in SVN)                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation; either version 2 of the    *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU Lesser General Public License for more details.                   *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef MULTIFRACTAL_ANALYZER_H
#define MULTIFRACTAL_ANALYZER_H

#include <map>
#include <vector>
/// Please note that this header is completely independent from the Q settings
/// You don't need to include "QSettings.h" if you with to use the analyzer,
/// only if you want to interpret the results!

#include "DecompositionLevel.h"

/// Final structure is a vector of data + its decomposition
/// Name based on Dynamic Multi-Fractal Analysis with Discrete Wavelet Transform
/// Suggested names like DMFADWT, diwtradymfrana... Too complex
/// Let's name it something simple
class MultifractalAnalyzer {
protected:

    // pointer to member type
    typedef DecompositionLevel* (MultifractalAnalyzer::* NewDataDWTFunctionPtr)(double);
    NewDataDWTFunctionPtr newDataDWTPtr;

    /// Updates the DWT decomposition, takes care of non-full buffer effects
    /// and automatically switched to the full buffer version when this happens
    DecompositionLevel* newDataDWTInitPhase(double newData);

    /// Faster version of the DWT update, assuming the buffer is full
    DecompositionLevel* newDataDWTFullRegime(double newData);
    int newDataDWTIdx;

    bool discardOlderData;
    bool monitorCorrelation;

    // These variables are only used when monitoring correlations
    double *sumOfWeightsInv, *correlationCoefficientDenominatorTerm1, *meanX;
    double* weights;
    double *squaredCorrelationSum;

    /// Protected helper called by the constructors
    void constructor(int nlevel, int ncoefmin, Wavelet* w, double* weights, int navg, bool discardOlderData, bool monitorCorrelation);

public:

    /// This is set to 2^L
    int nframes;

    /// These arrays are used only when averageAllFrames is true (the default)
    CyclicArrayTemplate<double*> hqHistory;

    /// These arrays are used only when both averageAllFrames & monitorCorrelation are true (not the default)
    CyclicArrayTemplate<double*> correlationHistory;

    /// Sum of local holder exponents estimates
    /// Size: 2^nlevel * ncoefmin
    /// => wait for reconstruction delay so exponents are estimated on all levels
    CyclicArray localHolderExponentsSum;
    /// Number of local holder exponents included in the sum
    /// => divide sum by num to get avg estimate over all frames of decomposition
    CyclicArrayTemplate<int> localHolderExponentsNum;

    /// Scale dependent constant for computing H(q) from q-power sums
    double **powerSumToHqConstant;

    /// Energy correction term for each Q and effective number of levels
    double **HQEnergyCorrection;

    /// Idem but for local Holder exponent estimation (-0.5 with uniform weighting scheme)
    double LHEnergyCorrection;

    /// Scale dependent constant for computing the b in y = b.s^(ax) weighted
    /// exponential fit from the a values (actually the log2 value of it)
    double **powerSumToExpFitResidualConstant;

    /// the h(q) values. Please review the QSettings.h file for the correspondance
    /// between the array index and the q value
    double *hq;

    /// squared correlation coefficient for the h(q) weighted exponential least square fitting.
    /// Available only when the 'monitorCorrelation' flag was used.
    /// Please review the QSettings.h file for the correspondance
    /// between the array index and the q value
    double *squaredCorrelation;

    int nlevel;

    /// In any method using a spectrum there is a delay to get low-frequency components.
    /// This is completely independent from this algorithm, but a general fact.
    /// In our case, the delay corresponds to the difference between the newly arrived point
    /// and the last point that can be reconstructed using the DWT.
    /// In practice, it is (2^L - 1)(wavelet->support - 2)
    int delay;

    DecompositionLevel *nextLevel;
    DecompositionLevel *nextLevelWaiting;

    /// Original data: the low pass for next level
    CyclicArray data;
    /// index of the last reconstructible point
    /// delay due to low pass frequency acquisition
    ///int delay;

    /// Analyzing wavelet
    Wavelet* wavelet;

    /** Public constructor takes the following arguments
        If you don't know what values to put, see suggestParametersForDataSize
        @param nlevel    Number of levels of decomposition
        @param ncoefmin  Size of the highest level of decomposition. For correct operation, this should be
                         at least max(2, wavelet->support - 2).
                         When keeping old data, this parameter has no effect on the result (hence the minimum is fine)
                         When discarding old data (non stationary series), set this parameter to some significant value
                         to make statistics on (10 may not be enough for example).
        @param w         The wavelet
        @param navg  Average that number of recent spectra in the final estimate. Set this to 1<<nlevel in order to average over all frames of decomposition. May be more, or less. Default is 1 (no averaging, only the current spectrum is used).
        @param discardOlderData   Whether expired data should be discarded or not. Default is true
        @param monitorCorrelation Whether to monitor the correlation coefficient or not (takes a % perf or so, strongly recommended to monitor the quality of the estimates). Default is true.

        @note  The data size is 2^nlevel * ncoefmin + (w->support - 2) * (2^nlevel - 1), with ^ = exponentiation
    */
    MultifractalAnalyzer(int nlevel, int ncoefmin, Wavelet* w, int navg=1, bool discardOlderData = true, bool monitorCorrelation = false);

    MultifractalAnalyzer(int nlevel, int ncoefmin, Wavelet* w, double* weights, int navg=1, bool discardOlderData = true, bool monitorCorrelation = false);

    /// Try to automatically adjust nlevel, ncoefmin, avg from the data size. Use this for fixed data set
    static void suggestParametersForDataSize(int dataSize, Wavelet* w, int& nlevel, int& ncoefmin, int& navg);

    ~MultifractalAnalyzer();

    /// Discrete Wavelet Transform the data into the next level
    /// low and high band coefficients
    void transform(CyclicArray& data, DecompositionLevel* level);

    /// restore data from next level by inverse DWT
    void restore(DecompositionLevel* level, CyclicArray& data);

    /// Update DWT for this new data.
    inline DecompositionLevel* newDataDWT(double newData) {
        return (this->*newDataDWTPtr)(newData);
    }

    /// Whether to compute the local holder estimates by using wavelet leaders
    /// over intervals K-1, K, K+1 or just over interval K, at each level.
    /// Default is false, use only K: less variance, but more bias.
    bool localHolderEstimatesUseLeaders3K;

    /// Feed a new data, update incrementally the DWT, and then optionally some values for the multifractal estimation method you wish to use.
    /// @param data The new data value
    /// @param updatePowerSums Whether to update the power sums for the Legendre transform method.
    /// @param updateLocalHolder Whether to update the local holder exponent estimates for the kernel density method
    /// @return 0 if there is not yet enough data to decompose into all levels
    /// @return 1 if all decomposition levels are used, but the latest level is not full yet
    /// @return 2 if all decomposition levels are used and full
    /// Note 1: You may also use newDataDWT if only the DWT update is needed
    /// Note 2: You can get early spectra estimates even during the initialization phase, though they may not be very significant. And even when the function has returned 2, you might still want to wait additional steps (to average over all frames for example)
    int newData(double data, bool updatePowerSums, bool updateLocalHolder);

    /// Get the multifractal spectrum, using the Legendre transform method.
    /// @param secondOrderLegendreApproximation When set (default) the series are first filtered for concavity, then a second-order estimate of the derivative is used for increased precision. This may occasionally return an empty map in degenerate cases where the series were really not concave. Not setting this value removes all checks, but you then risk getting erroneous artifacts.
    /// Note 1: h(q) values may be averaged over several frames for better precision
    /// Note 2: The result is only relevant when enough data has been collected, and the correct parameter was set when caling newData
    void getLegendreTransformSpectrum(std::map<double,double>& alpha_D, bool secondOrderLegendreApproximation = true);

    /// Get the multifractal spectrum, using the kernel density estimation of the local holder exponents method
    /// @param alpha The points at which the density is estimated -- OUTPUT parameter
    /// @param Dalpha The density values at each point -- OUTPUT parameter
    /// @param numAlpha The number of alpha values to return in the map. The default 0 means that this number is automatically selected from the number of available data, which may make the map quite big so you may wish to use lower counts to save time.
    /// @param kernelWidth The size of the kernel. The default value 0 means compute an optimal value automatically, but this is relatively computationally expensive. Alternatively, you might wish to call computeOptimalKernelWidth yourself, from time to time if not at every step, or provide another fixed kernel width if you're not satisfied with either the default or the computation time. Be careful not to under/oversmooth however...
    /// Note: The result is only relevant when enough data has been collected, and the correct parameter was set when caling newData
    /// @return the kernel width that was used, automatically determined if the argument was null
    double getLocalHolderSpectrum(std::vector<double>& alpha, std::vector<double>& Dalpha, int numAlpha = 0, double kernelWidth = 0.0, double epsilon = 1e-6);

    /// This function is called by the previous one once the alphas are fixed, but you may also call it directly if you wish to get D(alpha) for specific alpha values.
    /// @param alpha The points at which the density must be estimated -- INPUT parameter, unlike the previous function
    /// @param Dalpha The density values at the given points -- OUTPUT parameter
    /// @param kernelWidth The size of the kernel. The default value 0 means compute an optimal value automatically, but this is relatively computationally expensive. Alternatively, you might wish to call computeOptimalKernelWidth yourself, from time to time if not at every step, or provide another fixed kernel width if you're not satisfied with either the default or the computation time. Be careful not to under/oversmooth however...
    /// @return the kernel width that was used, automatically determined if the argument was null
    double estimateLocalHolderDensity(std::vector<double>& alpha, std::vector<double>& Dalpha, double kernelWidth = 0.0, double epsilon = 1e-6);

    /// Compute the AMISE-optimal kernel width for estimating the density of the current list of local holder exponents
    /// This routine relies on Vikas C. Raykar library for "Fast optimal bandwidth selection for kernel density estimation".
    /// Vikas C. Raykar and Ramani Duraiswami, In Proceedings of the sixth SIAM International Conference on Data Mining, Bethesda, April 2006, pp. 524-528. This gives a fast kernel derivative estimate that is exact at an arbitrary required precision.
    /// Then the Shether Jones Solve-the-equation plug-in method is applied on the kernel derivative in order to get the kernel width. "A reliable data-based bandwidth selection method for kernel density estimation", S.J. Sheather and M.C. Jones, Journal of the Royal Statistics Society B, 53:683-690, 1991
    /// The last step of the plug-in method requires solving a non-linear least squares problem.
    /// The Levenberg-Marquardt algorithm from lmfit / netlib / minpack is thus used to get the AMISE solution.
    /// @param epsilon Precision required on the solution. Default is 1e-6
    double computeOptimalKernelWidth(double epsilon = 1e-6);

    /// Use the histogram method, not the kernel method.
    /// @param alpha OUTPUT parameter, set to the bin centers
    /// @param histogram OUTPUT parameter, set to the bin counts
    /// @param numBins INPUT parameter, specify how many alphas to use. The bin size may be deduced from consecutive alphas.
    void getLocalHolderExponentsHistogram(std::vector<double>& alpha, std::vector<int>& histogram, int numBins);


    /// Use a band filter to reconstruct the latest possible data value with only the given frequency bands
    /// TODO: describe bands vector
    /// The reconstructed value is delay'ed by a fixed 'delay' amount
//    void setBandFilter(std::vector<int>& bands);
//    double filter();

    /// Note 1: The coefficients in each band are already available in the level->highPass
    ///       => Use that to make the time-frequency plot
    /// Note 2: when computing the multifractal spectrum, for Q=2, the power in the band is already computed



    /// Reset this analyzer. Use this instead of destroying/reallocating this object
    /// with the same parameters: the result is the same, but this function is much
    /// faster
    void reset();

    int effectiveNLevel;
};

#endif
