/***************************************************************************
 *   Authors:                                                              *
 *   - Nicolas Brodu                                                       *
 *                                                                         *
 *   Copyrights:                                                           *
 *   - Nicolas Brodu (initial author of incremfa)                          *
 *   - INRIA (modifications/diffs in SVN)                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation; either version 2 of the    *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU Lesser General Public License for more details.                   *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef QSETTINGS_H
#define QSETTINGS_H

#include <math.h>

/***************************************************************************
    These settings define the number of Q values, and their relation to the
    array indices.

    Please review these settings with attention.

    You should only need modify this file if you wish to change the Q
    settings.
 ***************************************************************************/

// Number of Q values, and array initializer
static const int NUM_Q = 64;
#define EMPTY_Q_INITIALIZER {0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.}

// Correspondance between array index and q value
// Here from -4 to +4, 0 excluded, with a step of 1/8
inline double GET_Q_FOR_INDEX(int i) {
    return ((i>=32) ? (31 - i) : (i + 1)) * 0.125;
}

// Define this flag if some of the Q values are negative
// See plusExponentiationNeg and minusExponentiationNeg
#define HAS_NEGATIVE_Q 1

// This routine adds up term^q, for q>0, to the power sums
inline void plusExponentiationPos(double term, double* powerSum) {
    double p0_5 = sqrt(term);
    double p0_25 = sqrt(p0_5);
    double p0_125 = sqrt(p0_25);
    double p0_375 = p0_125 * p0_25;
    double p0_625 = p0_125 * p0_5;
    double p0_75 = p0_25 * p0_5;
    double p0_875 = p0_375 * p0_5;
    powerSum[0] += p0_125;
    powerSum[1] += p0_25;
    powerSum[2] += p0_375;
    powerSum[3] += p0_5;
    powerSum[4] += p0_625;
    powerSum[5] += p0_75;
    powerSum[6] += p0_875;
    powerSum[7] += term;
    powerSum[8] += p0_125 * term;
    powerSum[9] += p0_25 * term;
    powerSum[10] += p0_375 * term;
    powerSum[11] += p0_5 * term;
    powerSum[12] += p0_625 * term;
    powerSum[13] += p0_75 * term;
    powerSum[14] += p0_875 * term;
    double sq = term * term;
    powerSum[15] += sq;
    powerSum[16] += p0_125 * sq;
    powerSum[17] += p0_25 * sq;
    powerSum[18] += p0_375 * sq;
    powerSum[19] += p0_5 * sq;
    powerSum[20] += p0_625 * sq;
    powerSum[21] += p0_75 * sq;
    powerSum[22] += p0_875 * sq;
    double tq = term * sq;
    powerSum[23] += tq;
    powerSum[24] += p0_125 * tq;
    powerSum[25] += p0_25 * tq;
    powerSum[26] += p0_375 * tq;
    powerSum[27] += p0_5 * tq;
    powerSum[28] += p0_625 * tq;
    powerSum[29] += p0_75 * tq;
    powerSum[30] += p0_875 * tq;
    powerSum[31] += term * tq;
}

// This routine subtracts term^q, for q>0 to the power sums
inline void minusExponentiationPos(double term, double* powerSum) {
    double p0_5 = sqrt(term);
    double p0_25 = sqrt(p0_5);
    double p0_125 = sqrt(p0_25);
    double p0_375 = p0_125 * p0_25;
    double p0_625 = p0_125 * p0_5;
    double p0_75 = p0_25 * p0_5;
    double p0_875 = p0_375 * p0_5;
    powerSum[0] -= p0_125;
    powerSum[1] -= p0_25;
    powerSum[2] -= p0_375;
    powerSum[3] -= p0_5;
    powerSum[4] -= p0_625;
    powerSum[5] -= p0_75;
    powerSum[6] -= p0_875;
    powerSum[7] -= term;
    powerSum[8] -= p0_125 * term;
    powerSum[9] -= p0_25 * term;
    powerSum[10] -= p0_375 * term;
    powerSum[11] -= p0_5 * term;
    powerSum[12] -= p0_625 * term;
    powerSum[13] -= p0_75 * term;
    powerSum[14] -= p0_875 * term;
    double sq = term * term;
    powerSum[15] -= sq;
    powerSum[16] -= p0_125 * sq;
    powerSum[17] -= p0_25 * sq;
    powerSum[18] -= p0_375 * sq;
    powerSum[19] -= p0_5 * sq;
    powerSum[20] -= p0_625 * sq;
    powerSum[21] -= p0_75 * sq;
    powerSum[22] -= p0_875 * sq;
    double tq = term * sq;
    powerSum[23] -= tq;
    powerSum[24] -= p0_125 * tq;
    powerSum[25] -= p0_25 * tq;
    powerSum[26] -= p0_375 * tq;
    powerSum[27] -= p0_5 * tq;
    powerSum[28] -= p0_625 * tq;
    powerSum[29] -= p0_75 * tq;
    powerSum[30] -= p0_875 * tq;
    powerSum[31] -= term * tq;
}

#ifdef HAS_NEGATIVE_Q

// Does for negative q what plusExponentiationPos does for positive q
// It's up to your program to decide how many array indices are allocated
// to positive and negative q
// In this example, with the same number of allocation, it's easy to
// re-use plusExponentiationPos
// For convenience this routine is passed term = 1/pos_term = pos_term^-1
inline void plusExponentiationNeg(double term, double* powerSum) {
    plusExponentiationPos(term, powerSum + 32);
}

// See plusExponentiationNeg
inline void minusExponentiationNeg(double term, double* powerSum) {
    minusExponentiationPos(term, powerSum + 32);
}

#endif

/***************************************************************************
    End of the Q dependent settings
 ***************************************************************************/

#endif
