/***************************************************************************
 *   Authors:                                                              *
 *   - Nicolas Brodu                                                       *
 *                                                                         *
 *   Copyrights:                                                           *
 *   - Nicolas Brodu (initial code from the incremfa project)              *
 *   - INRIA (modifications/diffs in SVN)                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation; either version 2 of the    *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU Lesser General Public License for more details.                   *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef WAVELET_H
#define WAVELET_H

#include "CyclicArray.h"

/** Orthogonal wavelet class. Base class for all Daubechies orthogonal wavelets
*/
class Wavelet {
protected:

    /// Reference counting
    int refCount;

    /// Base class constructor. See inherited classes
    Wavelet(int wsupport) : refCount(1), support(wsupport) {}

    /// Virtual destructor
    virtual ~Wavelet();

/// Helpers for subclasses. TODO: make getInstance() thread-safe
#define WAVELET_FACTORY_DEFINE(ClassName, support) \
protected: \
    ClassName() : Wavelet(support) {} \
    static ClassName* singleton; \
public: \
    static inline ClassName* getInstance() {return (singleton==0) ? (singleton=new ClassName()) : static_cast<ClassName*>(singleton->ref());} \
    virtual void convolve(CyclicArray& data, int idx, double* lphp) const; \
    virtual void reconstruct(CyclicArray& lp, CyclicArray& hp, int lpidx, int hpidx, double* evenodd) const; \
    virtual void smooth(CyclicArray& lp, int idx, double* evenodd) const; \
    virtual void delta(CyclicArray& hp, int idx, double* evenodd) const;

#define WAVELET_FACTORY_IMPLEMENT(ClassName) \
ClassName* ClassName::singleton = 0;

    
public:
    /// Size of the filter, number of coefficients
    int support;

    /// Ref counting functions. Warning: these are not thread-safe.
    inline Wavelet* ref() {++refCount; return this;}
    inline Wavelet* unref() {if (--refCount==0) {delete this; return 0;} return this;}

    /** convolves the data to produce low-pass (average) and high-pass (detail) coefficients
        @param data A cyclic array of data points to apply the filters to
        @param idx  The index in the data array for the start of the filter
        @param lphp A double[>=2] array that will hold the result.
                    The first (idx=0) element is the low-pass filter average,
                    the second (idx=1) is the high pass detail.
    */
    virtual void convolve(CyclicArray& data, int idx, double* lphp) const = 0;

    /// Reconstruct 2 data points from the given low-pass and high-pass coefficients
    /// The result buffer will hold the even and odd points, in this order
    inline void reconstruct(CyclicArray& lp, CyclicArray& hp, int idx, double* evenodd) {
        reconstruct(lp, hp, idx, idx, evenodd);
    }
    virtual void reconstruct(CyclicArray& lp, CyclicArray& hp, int lpidx, int hpidx, double* evenodd) const = 0;

    /// Only half the reconstruction: ignore high-pass coefficients
    /// This corresponds to "smoothing" the data
    virtual void smooth(CyclicArray& lp, int idx, double* evenodd) const = 0;

    /// Only half the reconstruction: ignore low-pass coefficients
    /// This can also be seen as delta = data - (smoothed version)
    virtual void delta(CyclicArray& hp, int idx, double* evenodd) const = 0;
};

/// Daubechies Wavelet with 4 coefficients
class DW4 : public Wavelet {
WAVELET_FACTORY_DEFINE(DW4, 4)
};

/// Daubechies Wavelet with 6 coefficients
class DW6 : public Wavelet {
WAVELET_FACTORY_DEFINE(DW6, 6)
};

/// Haar Wavelet
class HW : public Wavelet {
WAVELET_FACTORY_DEFINE(HW, 2)
};

#endif
