/***************************************************************************
 *   Authors:                                                              *
 *   - Nicolas Brodu                                                       *
 *                                                                         *
 *   Copyrights:                                                           *
 *   - INRIA                                                               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation; either version 2 of the    *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU Lesser General Public License for more details.                   *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <iostream>
#include <algorithm>
#include <fstream>
#include <sstream>
#include <vector>
#include <list>
#include <set>
using namespace std;
#include <math.h>

#ifndef QSETTINGS
#include "QSettings.h"
#else
#include QSETTINGS
#endif

#include "MultifractalAnalyzer.h"

const int default_nlevel = 6;
const string default_waveletName = "DW4";
const int default_ncoefmin = 10;

void help(const char* progname) {
cout<<
"This program computes the multifractal spectrum estimation of each variable in the input file, updated at each time value. The method used to estimate the multifractal spectrum is a Legendre transform of exponential fit of the power sums of the wavelet leaders.\n"
"\n"
"Usage: " << progname << " [options] input_filename output_filename\n"
"\n"
"The input file should consist of N variables per line, where each line is one observation in time.\n"
"The output file consists of X values derived from the multifractal spectrum depending on the options, for each variable, so there are N*X values per line.\n"
"\n"
"Options:\n"
"-L <int>: Number of decomposition levels (default=" << default_nlevel << ").\n"
"-B <bool,bool,...>: For each decomposition level, whether to include that level or not in the estimation. Example: If you know there is no relevant information in the first two bands use 0,0,1,1,..., this effectively subsamples the signal. You might also use this parameter to specify a range where the log-linear relation holds for estimating the exponents from the power sums. If you don't know what this is about, then the default 1,1,1... is probably fine anyway.\n"
"-W <string>: The wavelet to use (default="<<default_waveletName<<"). Available wavelets: DW4, DW6, HW. For Daubechies Wavelet 4 and 6, and Haar wavelet respectively.\n"
"-R <range_min,range_max,nsteps>: Compute nsteps values of D(alpha) where alpha ranges from range_min to range_max included (so nsteps should be at least 2). Note that in theory -infinity should be returned for values of alpha outside the holder exponent range present in the series. Here null values are returned for convenience. The default is to use -A and not -R, both options are incompatible.\n"
"-D <float,float,float...>: Each float specifies a percent of the maximum observed dimension (in practice dmax=1). The alpha values matching that level are then returned, which allows to reconstruct the spectrum. Example: 0.95,0.90 computes 5 alpha values: alpha_min(0.90*dmax), alpha_min(0.95*dmax), alpha(dmax), alpha_max(0.95*dmax), alpha_max(0.90*dmax), in this order. alpha_min and alpha_max respectively denote the smallest (resp. largest) alpha for which D(alpha) is equal to the given value. Use '-A 1' if you want to get only the maximum, in which case only one value is returned (this is the default).\n"
"-A <int>: Number of raw alphas to retain around the alpha for which D(alpha) is the maximum. Compared to the -D option, the alphas are here not interpolated between different values obtained from the log-linear fitting of the power sums. Neither the distance to the maximum in D(alpha) (-D option) nor the spacing in the alpha space (-R option) are fixed. Instead, the spacing between the power-sums (q exponents) is maintained.\n"
"-s <int>: Number of coefficients in the highest level of decomposition (default="<<default_ncoefmin<<").\n"
"-Z <int>: Suggest parameters for the given fixed data size.\n"
"-a: Keep all data points from the beginning of the series (default is to discard expired points for non-stationary series).\n"
"-w <int>: Moving average window: average results over the given number of previous computations (default=1: don't average).\n"
"-m <int>: Method to use for the spectrum estimation. 0: Local Holder exponent estimation (default). 1: Raw local holder exponents (no density estimation) 2: Legendre transform.\n"
"-k <float>: Specify a kernel width for the local Holder exponent density estimation. Use 0 to automatically estimate an 'optimal' width, tought be warned that for the present case a slight oversmoothing would be preferable as it reduces bias in the estimation of the alpha with maximal dimension (default=0). Recommendation: Use 0 first, note the value, and try to increase it to see if you get better results.\n"
"-n: Fill uncomputable values with NaN (default).\n"
"-0: Fill uncomputable values with 0.\n"
"-d: Do not output uncomputable values. Warning: the number of data values in the output file will differ from the input file.\n"
"-r: In case some series may not be computed at a given time step, reuse previous time values (default: fill with 0,nan, or stop).\n"
"-q: Quiet mode. Only critical messages are displayed.\n"
"-Q: Very quiet mode.\n"
"-S: Totally silent.\n"
"-t: Line number (time) from which to start writing the spectrum in the output file (default = 1, take all lines).\n"
"\n"
"Note: TODO: explain other methods this is for legendre method only. The options -D, -A and -R are incompatible. If you do not specify which method to use, '-D 1' is the default, which returns the alpha for which D(alpha) is maximum (that is, the top of the multifractal spectrum curve). This is also equivalently '-A 1'\n"
"Note 2: The range of q exponents is hardcoded and might not be sufficient to exhibit the full range of valid holder exponents. The -R option might thus return 0 values at points where non-null D(alpha) should be returned. More rarely, the maximum returned by the -A/-D options might not correspond to the real maximum if that value is outside the accessible range. A future extension might let the user specify the range of q exponents to overcome this limitation.\n"
<<flush;
}

vector<double> bands;

// prevent corruption by wavelet leaders from unused bands
void hackRemoveLeaderBand(DecompositionLevel *level, int l) {
    if (!level) return;
    if (l>1 && bands[l-1]!=0.0) return;
    if (l>1 && bands[l-1]==0.0) level->prevLevel=0;
    hackRemoveLeaderBand(level->nextLevel, l+1);
    hackRemoveLeaderBand(level->nextLevelWaiting, l+1);
}


int main(int argc, char** argv) {

    int quiet = 0;
    int nlevel = default_nlevel;
    string waveletName = default_waveletName;
    set<double> dmaxPercentLevels;
    dmaxPercentLevels.insert(1.0);
    double rmin=0.0, rmax=0.0;
    int rnstep=0;
    int ncoefmin = default_ncoefmin;
    int movingAvg = 1;
    bool keepOldPoints = false;
    int nalpha = 0;
    bool reuseMode = false;
    int suggestForSize = 0;
    bool useLegendreTransform = false;
    bool useLocalHolder = false;
    bool rawMode = false;
    int startTime = 1;
    double kernelWidthParam = 0.0;

    enum {FillNaN, FillZero, DoNotFill} fillMode;
    fillMode = FillNaN;

    int c; opterr = 0;
    while ((c=getopt(argc,argv,"L:B:W:R:D:A:s:Z:k:aw:n0dqQSrt:m:"))!=-1) switch(c) {
        case 'm': {
            int method = atoi(optarg);
            if (method==0) useLocalHolder = true;
            else if (method==1) {useLocalHolder = true; rawMode=true;}
            else if (method==2) useLegendreTransform = true;
            else {cout << "Invalid method specification: " << optarg << endl; help(argv[0]); return 0;}
            break;
        }
        case 'L': nlevel = atoi(optarg); break;
        case 'Z': suggestForSize = atoi(optarg); break;
        case 'B': {
            char *c = optarg;
            while (*c) {if (*c==',') *c= ' '; ++c;}
            stringstream tokenizer(optarg);
            int band;
            while (tokenizer >> band) bands.push_back((double)band);
            break;
        }
        case 'W': waveletName = optarg; break;
        case 'D': {
            char *c = optarg;
            while (*c) {if (*c==',') *c= ' '; ++c;}
            stringstream tokenizer(optarg);
            double percent;
            while (tokenizer >> percent) dmaxPercentLevels.insert(percent);
            break;
        }
        case 'A': nalpha = atoi(optarg); break;
        case 'R': {
            char *c = optarg;
            while (*c) {if (*c==',') *c= ' '; ++c;}
            stringstream tokenizer(optarg);
            tokenizer >> rmin;
            tokenizer >> rmax;
            tokenizer >> rnstep;
            if (rnstep<2) { cout << "Invalid number of steps" << endl; help(argv[0]); return 0; }
            if (rmax<rmin) { cout << "Invalid range specification" << endl; help(argv[0]); return 0; }
            break;
        }
        case 'k': kernelWidthParam = atof(optarg); break; // 0 => automatic => OK
        case 's': ncoefmin = atoi(optarg); break;
        case 'a': keepOldPoints = true; break;
        case 'w': movingAvg = atoi(optarg); break;
        case 'n': fillMode = FillNaN; break;
        case '0': fillMode = FillZero; break;
        case 'd': fillMode = DoNotFill; break;
        case 'q': quiet = 1; break;
        case 'Q': quiet = 2; break;
        case 'S': quiet = 10000; break;
        case 'r': reuseMode = true; break;
        case 't': {
            startTime = atoi(optarg); break;
            if (startTime<=0) {cout << "Invalid starting time " << optarg << endl; help(argv[0]); return 0; }
        }
        default: help(argv[0]); return 0;
    }

    Wavelet *wavelet = 0;
    if (waveletName=="DW4") wavelet = DW4::getInstance();
    else if (waveletName=="DW6") wavelet = DW6::getInstance();
    else if (waveletName=="HW") wavelet = HW::getInstance();
    else {
        cout << "Invalid wavelet name: " << waveletName << endl;
        help(argv[0]);
        return 0;
    }

    if (suggestForSize) {
        MultifractalAnalyzer::suggestParametersForDataSize(suggestForSize, wavelet, nlevel, ncoefmin, movingAvg);
        if (quiet<2) {
            cout << "Using wavelet " << waveletName<<", the suggested parameters for data size " << suggestForSize << " are: nlevel=" << nlevel << ", ncoefmin=" << ncoefmin<< ", movingAvg="<<movingAvg<<endl;
        }
    }


    if (useLegendreTransform && useLocalHolder) {
        cout << "Cannot specify both spectrum estimation methods with -m !" << endl;
        help(argv[0]);
        return 0;
    }

    // set default
    if (!useLegendreTransform && !useLocalHolder) useLocalHolder = true;

    if (movingAvg<=0) {
        cout << "Invalid number of elements for the moving average" << endl;
        help(argv[0]);
        return 0;
    }

    if (nlevel<=0) {
        cout << "Invalid number of levels" << endl;
        help(argv[0]);
        return 0;
    }

    if (nalpha<0) {
        cout << "Invalid number of alpha coefficients to retain" << endl;
        help(argv[0]);
        return 0;
    }

    if (bands.empty()) {
        bands.reserve(nlevel);
        for (int i=0; i<nlevel; ++i) bands.push_back(1.0);
    }

    if (bands.size()!=(unsigned int)nlevel) {
        cout << "Invalid number of bands" << endl;
        help(argv[0]);
        return 0;
    }
    bool bandsOK = true;
    for (int i=0; i<nlevel; ++i) if (bands[i]!=0.0 && bands[i]!=1.0) {bandsOK = false; break;}
    if (!bandsOK) {
        cout << "Invalid band filter:";
        for (int i=0; i<nlevel; ++i) cout << " " << bands[i];
        cout << endl;
        help(argv[0]);
        return 0;
    }

    int numCoefLowestValue = wavelet->support - 2;
    if (numCoefLowestValue < 2) numCoefLowestValue = 2;
    if (ncoefmin <= numCoefLowestValue) {
        cout << "Invalid number of elements for the highest level of decomposition" << endl;
        cout << "For the given wavelet, this should be at least " << numCoefLowestValue << endl;
        help(argv[0]);
        return 0;
    }


    // check D only if not in R / A mode
    if ((rnstep==0) && (nalpha==0)) {
        for (set<double>::const_iterator it = dmaxPercentLevels.begin(); it != dmaxPercentLevels.end(); ++it) {
            if (*it < 0.0 || *it>1.0) {
                cout << "Invalid dmax percentage specification - should be in 0..1" << endl;
                help(argv[0]);
                return 0;
            }
        }
    }

    if (argc - optind < 2) {
        help(argv[0]);
        return 0;
    }

    ifstream inputfile(argv[optind]);
    if (!inputfile) {
        cout << "Cannot read input file: " << argv[optind] << endl;
        return 0;
    }

    ofstream outputfile(argv[optind+1]);
    if (!outputfile) {
        cout << "Cannot write to output file: " << argv[optind+1] << endl;
        return 0;
    }

    if (quiet<10) {
        cout << "Processing file '" << argv[optind] << "' into '" << argv[optind+1] << "'" << endl;
        if (quiet<2) {
            cout << "Using wavelet " << waveletName << endl;
            cout << "With "<<nlevel<<" levels of decomposition" << endl;
            cout << "Band filter:";
            for (int i=0; i<nlevel; ++i) cout << " " << bands[i];
            cout << endl;
            if (nalpha>0) {
                if (nalpha==1) cout << "Computing alpha with maximum D(alpha)" << endl;
                else cout << "Retaining " << nalpha << " around the alpha with maximum D(alpha), included." << endl;
            }
            else if (rnstep==0) {
                if (dmaxPercentLevels.size()==1) cout << "Computing alpha with maximum D(alpha)" << endl;
                else {
                    cout << "Computing alpha values for D(alpha) levels, with levels at:";
                    for (set<double>::const_iterator it = dmaxPercentLevels.begin(); it != dmaxPercentLevels.end(); ++it) cout << " " << *it;
                    cout << endl;
                }
            }
            else {
                cout << "Computing " << rnstep << " D(alpha) values for alpha in the range [" << rmin << ".." << rmax<<"], bounds included." << endl;
            }

            switch (fillMode) {
                case FillNaN: cout << "Filling uncomputable values due to delay with NaN" << endl; break;
                case FillZero: cout << "Filling uncomputable values due to delay with 0" << endl;  break;
                case DoNotFill: cout << "Uncomputable values due to delay are not written to the output file" << endl; break;
            }
            if (movingAvg>1) cout << "Results are averaged over the last " << movingAvg << " computations" << endl;
            else cout << "Results are not averaged over successive computations" << endl;
        }
    }

    int num_output_values;
    if (nalpha!=0) num_output_values = nalpha;
    else if (rnstep!=0) num_output_values = rnstep;
    else num_output_values = dmaxPercentLevels.size() * 2 - 1;

    vector<MultifractalAnalyzer*> analyzers;

    string line; double value;
    vector<map<double,double> > prev_alpha_D;
    int lineNumber = 0;

    while (inputfile && !inputfile.eof()) {
        getline(inputfile, line);
        if (line.empty()) continue;
        stringstream linereader(line);
        vector<double> values;
        while (linereader >> value) {
            values.push_back(value);
        }
        if (values.empty()) continue; // should not happen

        if (prev_alpha_D.empty()) prev_alpha_D.resize(values.size());

        ++lineNumber;
        if (quiet<1) {
            cout << "Processing line " << lineNumber << " of the input file.";
            if (lineNumber < startTime) cout << " Skipping spectra computations." << endl;
            cout << endl;
        }

        // first time, create the analysers
        if (analyzers.empty()) {
            for (unsigned int i=0; i<values.size(); ++i) {
                analyzers.push_back(new MultifractalAnalyzer(nlevel, ncoefmin, wavelet, &bands[0], movingAvg,!keepOldPoints,false));
                // Horrible hack: prevent corruption by wavelet leaders from unused bands
                hackRemoveLeaderBand(analyzers.back()->nextLevel, 0);
                hackRemoveLeaderBand(analyzers.back()->nextLevelWaiting, 0);
            }
            if (quiet<2) {
                cout << "Sliding window over which computations are executed: " << analyzers[0]->data.capacity() << endl;
                cout << "Necessary delay due to low-frequency acquisition: " << analyzers[0]->delay << endl;
            }
        }

        // Feed data to the analysers
        bool donotfill=false;
        for (unsigned int i=0; i<values.size(); ++i) {
            int res = analyzers[i]->newData(values[i],useLegendreTransform,useLocalHolder);

            // Only take new data values into account
            if (lineNumber < startTime) continue;

            if (res<1) {
                if (fillMode!=DoNotFill) for (int ai=0; ai<num_output_values; ++ai) {
                    if (i+ai!=0) outputfile << " ";
                    if (fillMode==FillNaN) outputfile << "nan";
                    else if (fillMode==FillZero) outputfile << "0";
                } else donotfill = true;
                continue;
            }

            map<double,double> alpha_D;

            if (useLocalHolder) {
                if (res<2) continue;

                if (rawMode) {
                    for (int a=0; a<analyzers[i]->localHolderExponentsSum.size(); ++a) {
                        if ((i!=0) || (a!=0)) outputfile << " ";
                        outputfile << (analyzers[i]->localHolderExponentsSum[a] / analyzers[i]->localHolderExponentsNum[a]);
                    }
                    continue;
                }

                vector<double> alpha;
                vector<double> Dalpha;
                double kernelWidth = analyzers[i]->getLocalHolderSpectrum(alpha, Dalpha, 0, kernelWidthParam);
                if (quiet<1) {
                    cout << "kernelWidth = " << kernelWidth << endl;
                    cout << "H=[";
                    for (int a=0; a<analyzers[i]->localHolderExponentsSum.size(); ++a) {
                        cout << " " <<(analyzers[i]->localHolderExponentsSum[a] / analyzers[i]->localHolderExponentsNum[a]);
                    }
                    cout << "];"<<endl;
                    cout << "A=[";
                    for (unsigned int i=0; i<alpha.size(); ++i) {
                        cout << " " << alpha[i];
                    }
                    cout << "];" << endl << "D=[";
                    for (unsigned int i=0; i<Dalpha.size(); ++i) {
                        cout << " " << Dalpha[i];
                    }
                    cout << "];" << endl;
                }

                // hack. solution is to make API uniform
                for (unsigned int i=0; i<alpha.size(); ++i) alpha_D[alpha[i]] = Dalpha[i];
            }
            else {
                analyzers[i]->getLegendreTransformSpectrum(alpha_D);
            }

            {
                // empty?
                if (alpha_D.begin()==alpha_D.end()) {
                    if (reuseMode) {
                        if (quiet<2) cout << "Warning: could not get spectrum estimate at line " << lineNumber << " for series number " << (i+1) << ". Reusing previous time step computation." << endl;
                        alpha_D = prev_alpha_D[i];
                    }
                    if (alpha_D.begin()==alpha_D.end()) {
                        if (fillMode==DoNotFill) {
                            cerr << "Error: could not get spectrum estimate at line " << lineNumber << " for series number " << (i+1) << endl;
                            return 1;
                        }
                        if (quiet<2) {
                            cout << "Warning: could not get spectrum estimate at line " << lineNumber << " for series number " << (i+1) << ". Filling with ";
                            if (fillMode==FillNaN) cout << "nan";
                            else cout << "0";
                            cout << endl;
                        }
                        int count = 0;
                        if (nalpha > 0) count = nalpha;
                        else if (rnstep==0) count = dmaxPercentLevels.size() * 2 - 1;
                        else count = rnstep;
                        for (int a=0; a<count; ++a) {
                            if ((i!=0) || (a!=0)) outputfile << " ";
                            outputfile << ((fillMode==FillNaN)?"nan":"0");
                        }
                        continue;
                    }
                }

                prev_alpha_D[i] = alpha_D;
    /*
    if (i==0) {
    cout << "V: " << values[i] << endl;
                cout << "Corr: "; for (int j=0; j<NUM_Q; ++j) cout << analyzers[i]->squaredCorrelation[j] << " "; cout << endl;

                cout << "A=[";
                for(map<double,double>::const_iterator ait = alpha_D.begin(); ait!=alpha_D.end(); ++ait) {
                    cout << ait->first << " ";
                }
                cout <<"];"<<endl;
                cout << "D=[";
                for(map<double,double>::const_iterator ait = alpha_D.begin(); ait!=alpha_D.end(); ++ait) {
                    cout << ait->second << " ";
                }
                cout <<"];"<<endl;
    }
    */

                // raw alpha for fixed q values mode
                if (nalpha > 0) {

                    // Search for alpha with D_max. Dicho search on guaranteed convex function could be faster
                    double dmax = -1e300; int imax = -1; int curidx = -1;
                    vector<double> alphas;
                    for(map<double,double>::const_iterator ait = alpha_D.begin(); ait!=alpha_D.end(); ++ait) {
                        alphas.push_back(ait->first);
                        ++curidx;
                        if (ait->second > dmax) {
                            dmax = ait->second;
                            imax = curidx;
                        }
                    }

                    if (alphas.size()<(unsigned int)nalpha) {
                        alphas.resize((unsigned int)nalpha); // and zeros at the end
                    }
                    else {
                        int hmalpha = nalpha / 2;
                        int hpalpha = nalpha - hmalpha;
                        if (imax<hmalpha) imax = hmalpha;
                        alphas = vector<double>(alphas.begin()+(imax-hmalpha), alphas.begin()+(imax+hpalpha));
                    }

                    // Output result
                    if (lineNumber >= startTime) for (vector<double>::const_iterator ait = alphas.begin(); ait != alphas.end(); ++ait) {
                        if ((i!=0) || (ait!=alphas.begin())) outputfile << " ";
                        outputfile << *ait;
                    }

                }

                // alpha for given D(alpha) mode
                else if (rnstep==0) {

                    double dmax = -1e300; double alphamax = -1;
                    list<double> alphas;
                    for(map<double,double>::const_iterator ait = alpha_D.begin(); ait!=alpha_D.end(); ++ait) {
                        if (ait->second > dmax) {
                            dmax = ait->second;
                            alphamax = ait->first;
                        }
                    }
                    alphas.push_back(alphamax);

                    for (set<double>::const_reverse_iterator it = dmaxPercentLevels.rbegin(); it != dmaxPercentLevels.rend(); ++it) {
                        if (*it==1.0) continue; // alpha for dmax already computed
                        double amin=-1e300, amax=-1e300; bool bmin = false, bmax = false;
                        double dlevel = dmax * (*it);
                        map<double,double>::const_iterator ait = alpha_D.begin();
                        if (ait != alpha_D.end()) {
                            map<double,double>::const_iterator prevait = ait++;
                            for(; ait!=alpha_D.end(); ++ait) {
                                if (prevait->second < dlevel && ait->second >= dlevel) {
                                    amin = prevait->first + (dlevel - prevait->second) * (ait->first - prevait->first) / (ait->second - prevait->second);
                                    bmin = true;
                                }
                                if (prevait->second >= dlevel && ait->second < dlevel) {
                                    amax = prevait->first + (dlevel - prevait->second) * (ait->first - prevait->first) / (ait->second - prevait->second);
                                    bmax = true;
                                }
                                prevait = ait;
                            }
                        }
                        if (!bmin) amin = alpha_D.begin()->first;
                        if (!bmax) amax = alpha_D.rbegin()->first;
                        // sorted insert, since D_alpha is convex and dmaxPercentLevels is run in inverse sorted order
                        alphas.push_front(amin);
                        alphas.push_back(amax);
                    }

                    // Output result
                    for (list<double>::const_iterator ait = alphas.begin(); ait != alphas.end(); ++ait) {
                        if ((i!=0) || (ait!=alphas.begin())) outputfile << " ";
                        outputfile << *ait;
                    }
                }

                // D(alpha) for alpha in a given range mode
                else {
                    map<double,double>::const_iterator nextait = alpha_D.begin();
                    if (nextait == alpha_D.end()) for (int s = 0; s < rnstep; ++s) {if (i+s!=0) outputfile << " "; outputfile << 0;}
                    map<double,double>::const_iterator prevait = nextait++;
                    for (int s = 0; s < rnstep; ++s) {
                        double alpha = rmin + s * (rmax - rmin) / (rnstep - 1.0);
                        // increase the steps until we can have an alpha estimate
                        if (alpha < prevait->first) {
                            if (i+s!=0) outputfile << " ";
                            outputfile << 0;
                            continue;
                        }
                        // float comparison: very low probability this happens, but...
                        // covers the degenerate case where there is only one alpha
                        if (alpha == prevait->first) {
                            if (i+s!=0) outputfile << " ";
                            outputfile << prevait->second;
                            continue;
                        }
                        if (nextait == alpha_D.end()) {
                            if (i+s!=0) outputfile << " ";
                            outputfile << 0;
                            continue;
                        }
                        // prevait->first < alpha < nextait->first
                        if (alpha < nextait->first) {
                            if (i+s!=0) outputfile << " ";
                            outputfile << prevait->second + (nextait->second - prevait->second) * (alpha - prevait->first) / (nextait->first - prevait->first);
                            continue;
                        }
                        // prevait->first < nextait->first <= alpha
                        // restart with this alpha...
                        --s;
                        // .. but on the next interval
                        prevait = nextait++;
                    }
                }

            }
        }
        if (lineNumber >= startTime && !donotfill) outputfile << endl;
    }

    inputfile.close();

}

